import ast
import csv
import json
import random
from abc import ABC
from typing import Any, Callable, Iterator, Literal, Optional

from ._client import get_default_client
from .data import RELAISample


class Benchmark(ABC):
    """
    Abstract base class for defining and managing benchmarks.

    This class provides a foundational structure for benchmarks, enabling the
    download, and iteration of samples. It ensures that all concrete
    benchmark implementations have a unique identifier and a collection of samples
    to be used as inputs for AI agents and evaluators.

    Attributes:
        benchmark_id (str): A **unique identifier** for this specific benchmark.
        samples (list[RELAISample]): A list of `RELAISample` objects contained within this benchmark.
            Defaults to an empty list if not provided.
    """

    def __init__(self, benchmark_id: str, samples: Optional[list[RELAISample]] = None):
        """
        Args:
            benchmark_id (str): The unique identifier for the benchmark.
            samples (list[RELAISample], optional): A list of `RELAISample` objects to include
                in the benchmark. Defaults to an empty list.
        """
        self.benchmark_id = benchmark_id
        self.samples: list[RELAISample] = samples or []

    def __iter__(self) -> Iterator[RELAISample]:
        """
        Enables iteration over the samples within the benchmark as follows:

        ```python
        for sample in benchmark:
            # Process each sample
            pass
        ```

        Yields:
            RELAISample: Each `RELAISample` object contained in the benchmark.
        """
        for sample in self.samples:
            yield sample

    def __len__(self) -> int:
        """
        Returns the number of samples currently in the benchmark.

        Returns:
            int: The total count of `Sample` objects.
        """
        return len(self.samples)

    def sample(self, n: int = 1) -> list[RELAISample]:
        """
        Returns `n` random samples from the benchmark, with replacement.

        If `n` is greater than the total number of samples, samples may be
        repeated in the returned list.

        Args:
            n (int): The number of random samples to retrieve. Must be a positive integer.
                Defaults to 1.

        Returns:
            list[RELAISample]: A list containing `n` randomly selected `RELAISample` objects.

        Raises:
            ValueError: If `n` is less than or equal to 0.
        """
        if n <= 0:
            raise ValueError("Number of samples must be greater than 0.")
        return random.choices(self.samples, k=n)


class RELAIBenchmark(Benchmark):
    """
    A concrete implementation of `Benchmark` that downloads samples from the RELAI platform.

    Attributes:
        benchmark_id (str): The unique identifier (ID) of the RELAI benchmark to be loaded from the platform.
            You can find the benchmark ID in the metadata of the benchmark.
        samples (list[RELAISample]): A list of `RELAISample` objects contained within this benchmark.
    """

    def __init__(
        self,
        benchmark_id: str,
        field_name_mapping: Optional[dict[str, str]] = None,
        field_value_transform: Optional[dict[str, Callable]] = None,
        agent_input_fields: Optional[list[str]] = None,
        extra_fields: Optional[list[str]] = None,
    ):
        """
        Args:
            benchmark_id (str): The unique identifier for the RELAI benchmark.
                This ID is used to fetch the benchmark data from the RELAI platform.
            field_name_mapping (dict[str, str], optional): A mapping from field names
                returned by the RELAI API to standardized field names expected by the evaluators.
                If a field name is not present in this mapping, it is used as-is.
                Defaults to an empty dictionary.
            field_value_transform (dict[str, Callable], optional): A mapping from field names
                to transformation functions that convert field values from the RELAI API
                into the desired format. If a field name is not present in this mapping,
                the identity function is used (i.e., no transformation).
                Defaults to an empty dictionary.
            agent_input_fields (list[str], optional): A list of field names to extract from each
                sample for the `agent_inputs` dictionary. These fields are provided to the AI agent.
                Defaults to an empty list.
            extra_fields (list[str], optional): A list of field names to extract from each
                sample for the `extras` dictionary. These fields are also provided to the evaluators.
                Defaults to an empty list.
        """
        self._benchmark_id = benchmark_id
        self._field_name_mapping = field_name_mapping or {}
        self._field_value_transform = field_value_transform or {}

        self._agent_input_fields = agent_input_fields or []
        self._extra_fields = extra_fields or []
        super().__init__(benchmark_id=benchmark_id, samples=self.fetch_samples())

    def fetch_samples(self) -> list[RELAISample]:
        """
        Downloads samples from the RELAI platform and populates the `samples` attribute.

        This method fetches the benchmark data using the RELAI client and processes
        each sample to create `Sample` objects. The `samples` attribute is then
        updated with the newly fetched samples.
        """

        def strict_field_mapping(field: str) -> str:
            """
            Maps a field name returned by the RELAI API to a standardized name expected by the evaluators.

            Args:
                field (str): The field name as returned by the RELAI API.

            Returns:
                str: The standardized field name as defined in the `field_mapping` dictionary.
            """
            if field in self._field_name_mapping:
                return self._field_name_mapping[field]
            else:
                return field

        def strict_field_value_transform(field: str) -> Callable:
            """
            Returns a transformation function for a field value, if defined in `field_value_transform`.
            If no transformation is defined, it returns the identity function.

            Args:
                field (str): The field name for which the transformation is defined.

            Returns:
                Callable: A function that takes a value and applies the transformation.
            """
            if field in self._field_value_transform:
                return self._field_value_transform[field]
            return lambda x: x

        client = get_default_client()
        raw_samples = client.get_benchmark(self._benchmark_id)
        samples = []
        for raw_sample in raw_samples:
            try:
                agent_inputs = {
                    strict_field_mapping(field): strict_field_value_transform(field)(raw_sample[field])
                    for field in self._agent_input_fields
                }
            except KeyError as e:
                missing_key = e.args[0]
                raise ValueError(
                    f"""Expected agent input field `{missing_key}` is missing. """
                    f"""Make sure that the benchmark ID corresponds to a {self.__class__.__name__}."""
                )
            try:
                extras = {
                    strict_field_mapping(field): strict_field_value_transform(field)(raw_sample[field])
                    for field in self._extra_fields
                }
            except KeyError as e:
                missing_key = e.args[0]
                raise ValueError(
                    f"""Expected evaluation input field `{missing_key}` is missing. """
                    f"""Make sure that the benchmark ID corresponds to a {self.__class__.__name__}."""
                )
            sample = RELAISample(
                benchmark_id=self._benchmark_id,
                id=raw_sample["sample_uuid"],
                split=raw_sample["split"],
                agent_inputs=agent_inputs,
                extras=extras,
            )
            samples.append(sample)
        return samples


class RELAIQuestionAnsweringBenchmark(RELAIBenchmark):
    """
    A concrete implementation of `RELAIBenchmark` for question-answering tasks.
    All samples in this benchmark have the following fields:

    - `agent_inputs`:
        - `question`: The question to be answered by the AI agent.
    - `extras`:
        - `rubrics`:  A dictionary of rubrics for evaluating the answer.
        - `std_answer`: The standard answer to the question.
    }

    """

    def __init__(
        self,
        benchmark_id: str,
    ):
        """
        Args:
            benchmark_id (str): The unique identifier for the RELAI question-answering benchmark.
                This ID is used to fetch the benchmark data from the RELAI platform.
        """

        def rubric_transform(r: dict) -> dict:
            rubrics = {}
            for criterion in r["criteria"]:
                rubrics[criterion["description"]] = criterion["points"]
            return rubrics

        super().__init__(
            benchmark_id=benchmark_id,
            field_value_transform={"rubrics": rubric_transform},
            agent_input_fields=["question"],
            extra_fields=["rubrics", "std_answer"],
        )


class RELAISummarizationBenchmark(RELAIBenchmark):
    """
    A concrete implementation of `RELAIBenchmark` for summarization tasks. All samples in this benchmark have the following fields:

    - `agent_inputs`:
        - `source`: The text to be summarized.
    - `extras`:
        - `key_facts`: A list of key facts extracted from the source.
        - `style_rubrics`: A dictionary of rubrics for evaluating the style of the summary.
        - `format_rubrics`: A dictionary of rubrics for evaluating the format of the summary.
    """

    def __init__(
        self,
        benchmark_id: str,
    ):
        """
        Args:
            benchmark_id (str): The unique identifier for the RELAI summarization benchmark.
                This ID is used to fetch the benchmark data from the RELAI platform.
        """
        super().__init__(
            benchmark_id=benchmark_id,
            field_name_mapping={"source_text": "source", "source_keyfacts": "key_facts"},
            agent_input_fields=["source_text"],
            extra_fields=["source_keyfacts", "style_rubrics", "format_rubrics"],
        )


class RELAIAnnotationBenchmark(RELAIBenchmark):
    """
    A concrete implementation of `RELAIBenchmark` for benchmarks created from user annotations.
    All samples in this benchmark have the following fields:

    - `agent_inputs`:
        - The input(s) provided to the agent being evaluated.
    - `extras`:
        - `previous_outputs`: The previous outputs produced by the agent.
        - `desired_outputs`: The desired outputs as specified by the user.
        - `feedback`: The user feedback provided for the previous outputs.
        - `liked`: A boolean indicating whether the user liked the previous outputs.
    """

    def __init__(
        self,
        benchmark_id: str,
    ):
        """
        Args:
            client (AsyncRELAI): An instance of the AsyncRELAI client to interact with the RELAI platform.
            benchmark_id (str): The unique identifier for the RELAI summarization benchmark.
                This ID is used to fetch the benchmark data from the RELAI platform.
        """

        def nested_field_transform(value: str) -> Any:
            """
            Attempts to parse a string value as a nested field. If parsing fails, returns the original string.
            Args:
                value (str): The string value to be parsed.
            Returns:
                Any: The parsed JSON object, or the original string if parsing fails.
            """
            if value is None:
                return None
            try:
                return json.loads(value)
            except json.JSONDecodeError:
                try:
                    return ast.literal_eval(value)
                except (ValueError, SyntaxError):
                    return value

        def liked_transform(value: bool) -> Literal["like", "dislike"]:
            if value is None:
                return "not provided"
            return "like" if value else "dislike"

        super().__init__(
            benchmark_id=benchmark_id,
            field_name_mapping={
                "agent_outputs": "previous_outputs",
                "serialized_simulation_config": "simulation_config",
                "agent_inputs": "all_inputs",
            },
            agent_input_fields=["agent_inputs"],
            extra_fields=["agent_outputs", "desired_outputs", "feedback", "liked", "serialized_simulation_config"],
            field_value_transform={
                "agent_outputs": nested_field_transform,
                "desired_outputs": nested_field_transform,
                "serialized_simulation_config": nested_field_transform,
                "agent_inputs": nested_field_transform,
                "liked": liked_transform,
            },
        )


class CSVBenchmark(Benchmark):
    """
    A concrete implementation of `Benchmark` that loads samples from a CSV file.

    Attributes:
        benchmark_id (str): The unique identifier (ID) of the benchmark to be loaded from the CSV file.
            Defaults to the CSV file name.
        samples (list[Sample]): A list of `Sample` objects contained within this benchmark.
    """

    def __init__(
        self,
        csv_file: str,
        agent_input_columns: Optional[list[str]] = None,
        extra_columns: Optional[list[str]] = None,
        benchmark_id: Optional[str] = None,
    ):
        """
        Args:
            csv_file (str): The path to the CSV file containing benchmark samples.
            agent_input_columns (list[str], optional): A list of column names in the CSV file
                that should be used as inputs for the AI agent. Defaults to an empty list.
            extra_columns (list[str], optional): A list of column names in the CSV file
                that could be used as inputs for evaluators. Defaults to an empty list.
            benchmark_id (str, optional): A unique identifier for the benchmark.
                If not provided, it defaults to the name of the CSV file.
        """
        benchmark_id = benchmark_id or csv_file
        agent_input_columns = agent_input_columns or []
        extra_columns = extra_columns or []
        samples = []
        with open(csv_file, encoding="utf-8-sig") as csvfile:
            reader = csv.DictReader(csvfile)
            for idx, row in enumerate(reader):
                sample = RELAISample(
                    benchmark_id=benchmark_id,
                    id=str(idx),
                    agent_inputs={col: self._deserialize_str(row[col]) for col in agent_input_columns},
                    extras={col: self._deserialize_str(row[col]) for col in extra_columns},
                )
                samples.append(sample)
        super().__init__(benchmark_id=benchmark_id, samples=samples)

    def _deserialize_str(self, value: str) -> Any:
        """
        Attempts to deserialize a string value into its original type.
        This is useful for converting string representations of complex types back to their original form.
        """
        try:
            return json.loads(value)
        except json.JSONDecodeError:
            # If deserialization fails, return the original string
            return value



from ._client import RELAI, AsyncRELAI
from ._exceptions import RELAIError
from .data import AgentInputs, AgentLog, AgentOutputs, CriticoLog, EvaluatorLog, RELAISample, SimulationTape
from .simulator import AsyncSimulator, SyncSimulator, get_current_simulation_mode, random_env_generator, simulated

__all__ = [
    "AgentLog",
    "AgentInputs",
    "AgentOutputs",
    "AsyncRELAI",
    "AsyncSimulator",
    "CriticoLog",
    "EvaluatorLog",
    "get_current_simulation_mode",
    "random_env_generator",
    "RELAI",
    "RELAIError",
    "RELAISample",
    "SimulationTape",
    "simulated",
    "SyncSimulator",
]



from contextlib import contextmanager
from contextvars import ContextVar


class RELAITrackingFlag:
    """
    A simple mutable boolean-like flag object.

    Useful for controlling state across different parts of a system, such as toggling
    a global tracking mode.
    """

    def __init__(self, init: bool):
        """
        Initializes the flag with a given value.

        Args:
            init (bool): Initial value of the flag.
        """
        self._value = init

    def __bool__(self) -> bool:
        """
        Returns the boolean value of the flag.

        Returns:
            bool: The current value of the flag.
        """
        return self._value

    def set(self, value: bool):
        """
        Sets the flag to a new value.

        Args:
            value (bool): The new value to assign to the flag.
        """
        self._value = value


relai_tracking_flag = RELAITrackingFlag(False)  # for root context
relai_tracking_flag_var: ContextVar[RELAITrackingFlag] = ContextVar("relai_tracking_flag", default=relai_tracking_flag)


@contextmanager
def set_current_tracking(flag: RELAITrackingFlag):
    """
    A context manager to set the tracking flag for the enclosed context.
    """
    token = relai_tracking_flag_var.set(flag)
    try:
        yield
    finally:
        relai_tracking_flag_var.reset(token)


def get_current_tracking() -> RELAITrackingFlag:
    """
    Returns the tracking flag instance for the current execution context.
    """
    return relai_tracking_flag_var.get()


def tracking_on():
    """
    Enables tracking by setting the global `relai_tracking_flag` to True.
    """
    relai_tracking_flag = get_current_tracking()
    relai_tracking_flag.set(True)


def tracking_off():
    """
    Disables tracking by setting the global `relai_tracking_flag` to False.
    """
    relai_tracking_flag = get_current_tracking()
    relai_tracking_flag.set(False)



import asyncio
from collections.abc import Callable, Sequence
from functools import cached_property
from typing import Any, ClassVar, cast, overload
from uuid import uuid4

from agents import Agent, Runner, SQLiteSession, Tool, function_tool
from agents.extensions.models.litellm_model import LitellmModel

from .._client import get_default_client
from ..utils import log_persona, no_trace
from .base_mocker import BaseMocker


class Persona(BaseMocker):
    """
    A mocker class for creating AI personas with specific behaviors and tools.
    A persona can be used to mimic a particular role by defining a system prompt and optionally equipping it with
    tools.

    Attributes:
        user_persona (str): The description of the persona's characteristics and behavior.
        intent (str | None): The intent or goal of the persona during interactions.
        starting_message (str | None): An optional initial message that the persona will use to start interactions.
        tools (Optional[list[Tool]]): A list of tools that the persona can use.
    """

    prompt_template: ClassVar[str] = (
        """You are an assistant helping with testing another AI agent. Your job is to """
        """impersonate a user and interact with the agent based on the following """
        """persona description. The intent of the user is also provided below. Only respond """
        """with what a user would say, without any extraneous phrases like "Here's my prompt".\n"""
        """<intent>{intent}</intent>\n"""
        """<persona>{persona}</persona>\n"""
    )

    additional_prompt_trajectory: ClassVar[str] = (
        """Additionally, you should repeat the exact behavior of {identifier_in_trajectory} """
        """in the following trajectory provided as long as it seems appropriate in the current context. """
        """The goal is to try recreating the same situation recorded in the trajectory.\n"""
        """<trajectory>\n{trajectory}\n</trajectory>\n"""
    )

    def __init__(
        self,
        user_persona: str,
        intent: str | None = None,
        starting_message: str | None = None,
        trajectory: str | None = None,
        identifier_in_trajectory: str = "user",
        tools: list[Callable[..., Any]] | None = None,
        model: str | LitellmModel = "gpt-5-mini",
    ):
        """
        Initializes the Persona with a description, intent, optional starting message, tools, and model.

        Args:
            user_persona (str): The description of the persona's characteristics and behavior.
            intent (str | None): The intent or goal of the persona during interactions.
            starting_message (str | None): An optional initial message that the persona will use to start interactions.
            trajectory (str | None): If provided, the persona will try to mimic the behavior of a certain entity
                (defaults to "user") recorded in the trajectory. The entity to mimic can be configured through
                `identifier_in_trajectory`.
            identifier_in_trajectory (str): A string to identify the entity to mimic when a trajectory is provided.
                Defaults to "user".
            tools (Optional[list[Tool]]): A list of tools that the persona can use.
            model (str | LitellmModel): The AI model to use for simulating the persona's behavior.
                This can be a string identifier for OpenAI models (e.g. gpt-5-mini) or a LitellmModel
                (from agents.extensions.models.litellm_model import LitellmModel). For a full list of
                models supported in LiteLLM, see https://docs.litellm.ai/docs/providers
        """
        super().__init__()
        self.name = f"persona-{uuid4().hex}"
        self.user_persona = user_persona
        self.intent = intent
        self.starting_message = starting_message
        self.trajectory = trajectory
        self.identifier_in_trajectory = identifier_in_trajectory
        tools = tools or []
        self.tools: list[Tool] = [function_tool(tool) for tool in tools]
        self.model = model
        self._session = SQLiteSession(self.name)
        self._start = True

    @cached_property
    def agent(self) -> Agent:
        instructions = self.prompt_template.format(intent=self.intent or self._func_doc, persona=self.user_persona)
        if self.trajectory is not None:
            instructions += self.additional_prompt_trajectory.format(
                identifier_in_trajectory=self.identifier_in_trajectory, trajectory=self.trajectory
            )

        return Agent(
            name=self.name,
            tools=self.tools,
            instructions=instructions,
            model=self.model,
            output_type=self.output_type,
        )

    async def _add_starting_message(self, agent_input: str) -> None:
        await self._session.add_items(
            [
                {"role": "user", "content": agent_input},
                {"role": "assistant", "content": cast(str, self.starting_message)},
            ]
        )

    def _run(self, *args, **kwargs):
        agent_input = str(
            {
                "args": args,
                "kwargs": kwargs,
            }
        )

        if self._start and self.starting_message is not None:
            asyncio.run(self._add_starting_message(agent_input))
            self._start = False
            output = self.starting_message
        else:
            with no_trace():
                result = Runner.run_sync(
                    self.agent,
                    agent_input,
                    session=self._session,
                )
            output = result.final_output

        log_persona(
            f"RELAI Persona ({self.name})",
            self.model if isinstance(self.model, str) else self.model.model,
            agent_input,
            output,
            note=f"Persona id: {self.name}",
        )
        return output

    async def _arun(self, *args, **kwargs):
        agent_input = str(
            {
                "args": args,
                "kwargs": kwargs,
            }
        )

        if self._start and self.starting_message is not None:
            await self._add_starting_message(agent_input)
            self._start = False
            output = self.starting_message
        else:
            with no_trace():
                result = await Runner.run(
                    self.agent,
                    agent_input,
                    session=self._session,
                )
            output = result.final_output

        log_persona(
            f"RELAI Persona ({self.name})",
            self.model if isinstance(self.model, str) else self.model.model,
            agent_input,
            output,
            note=f"Persona id: {self.name}",
        )
        return output

    def serialize(self) -> dict[str, str]:
        return {
            "user_persona": self.user_persona,
            "intent": self.intent or "",
            "starting_message": self.starting_message or "",
        }


class PersonaSet(Sequence[Persona]):
    """
    A collection of Persona instances loaded from a persona set on the RELAI platform.
    """

    def __init__(self, persona_set_id: str, **persona_kwargs: Any) -> None:
        """
        Initializes the PersonaSet with the given persona set ID.

        Args:
            persona_set_id (str): The ID of the persona set on the RELAI platform.
            **persona_kwargs: Keyword arguments that are forwarded to each Persona created from the set.
        """
        self.persona_set_id = persona_set_id
        self._user_personas = None
        self._personas = None
        self._persona_kwargs = persona_kwargs

    def user_personas(self) -> list[str]:
        if self._user_personas is None:
            # Lazy load the personas from the platform
            client = get_default_client()
            self._user_personas = client.get_persona_set(self.persona_set_id)
        return self._user_personas

    def personas(self) -> list[Persona]:
        if self._personas is None:
            self._personas = [
                Persona(user_persona=persona, **self._persona_kwargs) for persona in self.user_personas()
            ]
        return self._personas

    @overload
    def __getitem__(self, index: int) -> Persona: ...

    @overload
    def __getitem__(self, index: slice) -> list[Persona]: ...

    def __getitem__(self, index: int | slice) -> Persona | list[Persona]:
        return self.personas()[index]

    def __len__(self) -> int:
        return len(self.personas())



import asyncio
import os
import time
from abc import ABC, abstractmethod
from typing import Any, Optional

import aiohttp
import httpx

from ._exceptions import RELAIError
from .schema.visual import ConfigOptVizSchema, GraphOptVizSchema


class BaseRELAI(ABC):
    def __init__(self, api_key: Optional[str] = None, api_url: Optional[str] = None):
        if api_key is None:
            api_key = os.getenv("RELAI_API_KEY")
        self.api_key = api_key
        if api_url is None:
            api_url = os.getenv("RELAI_API_URL", "https://api.relai.ai")
        self.api_url = api_url
        self._client = None

    @property
    def headers(self) -> dict[str, str]:
        if self.api_key is None:
            raise RELAIError(
                (
                    """API key must be provided either as an argument or through the `RELAI_API_KEY` environment """
                    """variable. An API key is required for all interactions with the RELAI Platform."""
                )
            )
        return {
            "Content-Type": "application/json",
            "Authorization": f"Token {self.api_key}",
            "Accept-Encoding": "gzip, deflate",
        }

    @property
    @abstractmethod
    def client(self) -> Any:
        raise NotImplementedError("Subclasses must implement the 'client' property.")


class RELAI(BaseRELAI):
    """
    A client for interacting with the RELAI Platform.

    Args:
        api_key (Optional[str]): Your RELAI API key. You can find this on RELAI Platform:
            Enterprise > Access Management > API Keys. API key can also be set
            as an environment variable `RELAI_API_KEY`.
        api_url (Optional[str]): The base URL for the RELAI API. Defaults to "https://api.relai.ai".
            API URL can also be set as an environment variable `RELAI_API_URL`.

    Raises:
        RELAIError: If the API key is not provided or cannot be found in the environment variables.
    """

    def __init__(self, api_key: Optional[str] = None, api_url: Optional[str] = None):
        super().__init__(api_key=api_key, api_url=api_url)

    @property
    def client(self) -> httpx.Client:
        if self._client is None:
            self._client = httpx.Client(base_url=self.api_url, headers=self.headers)
        return self._client

    def close(self):
        if self._client is not None:
            self._client.close()
            self._client = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.client.close()

    def _request(self, method: str, url: str, **kwargs: Any) -> Any:
        """Performs an HTTP request and handles common errors."""
        try:
            response = self.client.request(method, url, **kwargs)
            response.raise_for_status()
            return response.json()
        except httpx.HTTPStatusError as e:
            # Re-raise with more context from the response body if possible
            raise RELAIError(f"HTTP error occurred: {e.response.status_code} - {e.response.text}") from e
        except httpx.RequestError as e:
            raise RELAIError(f"Request failed: {e}") from e

    def _get(self, *args, **kwargs) -> Any:
        return self._request("GET", *args, **kwargs)

    def _post(self, *args, **kwargs) -> Any:
        return self._request("POST", *args, **kwargs)

    def _put(self, *args, **kwargs) -> Any:
        return self._request("PUT", *args, **kwargs)

    def is_connected(self) -> bool:
        """Checks if the client can connect to the RELAI API."""

        try:
            self._get("health/")
        except RELAIError:
            return False
        else:
            return True

    def get_benchmark_keywords(self) -> list[str]:
        """
        Retrieves the list of benchmark keywords from the RELAI API.

        Returns:
            list[str]: A list of benchmark keywords.
        """
        response = self._get("api/v1/benchmark/keywords/")
        return response

    def get_benchmark(self, benchmark_id: str) -> list[dict[str, Any]]:
        """
        Retrieves the benchmark data for a given benchmark ID.

        Args:
            benchmark_id (str): The ID of the benchmark to retrieve.

        Returns:
            list[dict[str, Any]]: Samples from the benchmark.

        Raises:
            RELAIError: If the request fails or the benchmark is not found.
        """
        details_url = f"api/v1/benchmarks/all/{benchmark_id}/"

        metadata = self._get(details_url)

        samples_data = self._get(f"api/v1/benchmarks/all/{benchmark_id}/samples/", params={"no_pagination": "true"})

        samples = []
        for s in samples_data:
            split = s.get("split", "All").lower().capitalize()
            sample = {
                "sample_uuid": s["sample_uuid"],
                "split": split,
                **metadata.get("hyperparameters", {}),  # Fields like `style_rubrics`, `format_rubrics`
                **s["column_values"],
            }
            samples.append(sample)
        return samples

    def get_evaluator_response(self, evaluator_name: str, benchmark_id: str, sample_id: str, **kwargs: Any) -> Any:
        """
        Calls a RELAI evaluator with the specified name and parameters.

        Args:
            evaluator_name (str): The name of the evaluator to call.
            **kwargs: Additional parameters to pass to the evaluator.

        Returns:
            Any: The response from the evaluator.

        Raises:
            RELAIError: If the request fails or the evaluator is not found.
        """
        payload = {"evaluator_type": evaluator_name, "benchmark_id": benchmark_id, "sample_id": sample_id, **kwargs}
        response = self._post("api/v1/enterprise/evaluator/", json=payload)
        return response["evaluation_result"]["response"]

    def _execute_maestro_task(self, url: str, data: dict, timeout: int = 1800) -> dict[str, Any]:
        """Executes a Maestro task by POSTing data and polling for the result."""
        response = self._post(url, json=data)
        task_id = response["result"]["task_id"]
        return self.poll_maestro_task(task_id, timeout=timeout)

    def poll_maestro_task(self, task_id: str, timeout: int = 1800) -> Any:
        """
        Poll the status of a Maestro until it is complete or there is an error.
        If and when the task is complete, return the output.

        Args:
            task_id (str): The ID of the Maestro task to poll.
            timeout (int): The maximum time to wait for the task to complete (in seconds).

        Returns:
            Any: The output of the Maestro task if it is complete.

        Raises:
            RELAIError: If the polling request fails.
        """
        payload = {"task_id": task_id}
        sleep_interval = 0.25  # in seconds
        start = time.time()

        while True:
            elapsed = time.time() - start
            if elapsed > timeout:
                raise RELAIError(f"Maestro task timed out after {timeout:.2f} seconds")
            response = self._post("api/v1/maestro/poll-task/", json=payload)
            status = response["result"]["status"]
            if status == "PENDING":
                time.sleep(sleep_interval)
                sleep_interval = min(sleep_interval * 2, 10)  # Exponential backoff, max 10 seconds
                continue
            elif status == "SUCCESS":
                return response["result"]
            else:
                raise RELAIError(f"Maestro task failed: {status}")

    def process_test_case(self, data: dict) -> tuple[list[str], dict[str, int]]:
        """
        Process a new test case using the RELAI platform for later selecting samples for optimization.

        Args:
            data (dict): The data required for processing the test case.

        Returns:
            list[str]: A list of tags for the new test case.
            dict[str, int]: A dictionary of active tags and tag counts.

        Raises:
            RELAIError: If the processing request fails.
        """
        response = self._execute_maestro_task("api/v1/maestro/process-test-case/", data)
        tags = response["response"]["tags"]
        updated_active_tags = response["response"]["updated_active_tags"]

        return tags, updated_active_tags

    def apply_structure_to_code(self, data: dict) -> str:
        """
        Applies the suggested structural changes to the agent code with RELAI platform.

        Args:
            data (dict): The data required for applying structural changes to the agent code.

        Returns:
            str: The updated agent code after applying the suggested structural changes.

        Raises:
            RELAIError: If the application request fails.
        """
        response = self._execute_maestro_task("api/v1/maestro/apply-structure-to-code/", data)

        return response["response"]["code"]

    def optimize_structure(self, data: dict) -> str:
        """
        Optimizes the structure of an agent using the RELAI platform.

        Args:
            data (dict): The data required for structure optimization.

        Returns:
            str: The suggestion for optimizing the agent's structure.

        Raises:
            RELAIError: If the optimization request fails.
        """
        response = self._execute_maestro_task("api/v1/maestro/optimize-structure/", data)
        return response["response"]["suggestion"]

    def propose_values(self, data: dict) -> tuple[str, dict[str, str]]:
        """
        Proposes new values for the agent's parameters.

        Args:
            data (dict): The data required for proposing new values.

        Returns:
            tuple[str, dict[str, str]]: A tuple containing the analysis of agent implementation
                and the proposed parameter values.

        Raises:
            RELAIError: If the proposal request fails.
        """
        response = self._execute_maestro_task("api/v1/maestro/propose-values/", data)
        analysis = response["response"]["analysis"]
        proposed_values = response["response"]["proposed_values"]
        return analysis, proposed_values

    def review_values(self, data: dict) -> dict[str, Any]:
        """
        Reviews proposed values for the agent's parameters.

        Args:
            data (dict): The data required for reviewing proposed values.

        Returns:
            dict[str, Any]: The review of the proposed values.

        Raises::
            RELAIError: If the review request fails.
        """
        response = self._execute_maestro_task("api/v1/maestro/review-values/", data)
        return response["response"]["review"]

    def select_version(self, data: dict) -> int:
        """
        Selects the version to evaluate or apply

        Args:
            data (dict): The data to use for selecting the version.

        Returns:
            int: The ID of the selected version.

        Raises:
            RELAIError: If the selection request fails.
        """
        response = self._execute_maestro_task("api/v1/maestro/select-version/", data)
        return response["response"]["version"]

    def upload_trace(self, data: dict[str, Any]) -> Any:
        """
        Uploads OpenInference execution logs of a run to the RELAI platform.

        Args:
            data (dict[str, Any]): A dictionary containing execution log data to upload.

        Returns:
            str: The id for the uploaded execution logs.

        Raises:
            RELAIError: If the upload request fails.
        """
        response = self._post("/api/v1/traces/runs/", json=data)
        return response["trace_uuid"]

    def upload_run(
        self,
        group_id: str,
        trace_id: str,
        agent_inputs: dict[str, Any] | None = None,
        agent_outputs: dict[str, Any] | None = None,
        extras: dict[str, Any] | None = None,
        serialized_simulation_config: dict[str, Any] | None = None,
        agent_version_uuid: str | None = None,
        environment_uuid: str | None = None,
    ) -> None:
        agent_inputs = agent_inputs or {}
        agent_outputs = agent_outputs or {}
        extras = extras or {}
        serialized_simulation_config = serialized_simulation_config or {}
        self._post(
            "api/v1/agents/agent-runs/",
            json={
                "group_id": group_id,
                "trace": trace_id,
                "agent_inputs": agent_inputs,
                "agent_outputs": agent_outputs,
                "extras": extras,
                "serialized_simulation_config": serialized_simulation_config,
                "agent_version_uuid": agent_version_uuid,
                "environment_uuid": environment_uuid,
            },
        )

    def upload_critico_log(
        self,
        trace_id: str,
        evaluator_logs: list[dict[str, Any]],
        aggregate_score: float,
        aggregate_feedback: str,
    ) -> None:
        """
        Submits evaluation data to the RELAI platform.

        Args:
            evaluation_data (dict): The evaluation data to submit.

        Returns:
            Any: The response from the submission.

        Raises:
            RELAIError: If the submission fails.
        """
        self._post(
            "api/v1/agents/critico-logs/",
            json={
                "trace": trace_id,
                "evaluator_logs": evaluator_logs,
                "aggregate_score": aggregate_score,
                "aggregate_feedback": aggregate_feedback,
            },
        )

    def upload_critico_report(self, title: str, raw_evaluation_sets: list[dict[str, Any]]) -> None:
        """
        Submit a Critico report to the RELAI platform.

        Args:
            title (str): The title of the Critico report.
            raw_evaluation_sets (list[dict[str, Any]]): The data of the Critico report.
        """
        self._post(
            "/api/v1/benchmarks/evaluations/submit/",
            json={
                "title": title,
                "raw_evaluation_sets": raw_evaluation_sets,
            },
        )

    def update_config_opt_visual(self, config_viz: ConfigOptVizSchema, uuid: str | None = None) -> str:
        """
        Updates the configuration optimization visualization data on the RELAI platform.

        Args:
            config_viz (ConfigOptVizSchema): The configuration optimization visualization data to update.
            uuid (str, optional): The UUID of the configuration optimization visualization to update. If not provided, a new visualization will be created.

        Returns:
            str: The uuid of the updated or created configuration optimization visualization.

        Raises:
            RELAIError: If the update request fails.
        """

        if uuid is None:
            response = self._post("api/v1/maestro/config-optimizer/", json=config_viz.model_dump())
            return response["config_opt_viz_util_uuid"]
        else:
            self._put(f"api/v1/maestro/config-optimizer/{uuid}/", json=config_viz.model_dump())
            return uuid

    def update_graph_opt_visual(self, graph_vis: GraphOptVizSchema, uuid: str | None = None) -> str:
        """
        Updates the graph optimization visualization data on the RELAI platform.

        Args:
            graph_vis (GraphOptVizSchema): The graph optimization visualization data to update.
            uuid (str, optional): The UUID of the graph optimization visualization to update. If not provided, a new visualization will be created.

        Returns:
            str: The uuid of the updated or created graph optimization visualization.

        Raises:
            RELAIError: If the update request fails.
        """

        if uuid is None:
            response = self._post("api/v1/maestro/graph-optimizer/", json=graph_vis.model_dump())
            return response["graph_opt_viz_util_uuid"]
        else:
            self._put(f"api/v1/maestro/graph-optimizer/{uuid}/", json=graph_vis.model_dump())
            return uuid

    def get_persona_set(self, persona_set_id: str) -> list[str]:
        """
        Retrieves the details of a persona set by its ID.

        Args:
            persona_set_id (str): The ID of the persona set to retrieve.

        Returns:
            list[str]: A list of system prompts in the persona set.
        """
        response = self._get(f"api/v1/persona/persona-sets/{persona_set_id}/")
        return response["system_prompts"]

    def get_custom_evaluator(self, evaluator_id: str) -> dict[str, Any]:
        """
        Retrieves the details of a custom evaluator by its ID.

        Args:
            evaluator_id (str): The ID of the evaluator to retrieve.

        Returns:
            dict[str, Any]: A dictionary containing the name and parameters of the evaluator.
        """
        response = self._get(f"api/v1/benchmarks/functions/{evaluator_id}/")
        return {"name": response["name"], "parameters": response["parameters"]}


class AsyncRELAI(BaseRELAI):
    """
    A client for interacting with the RELAI Platform.

    Args:
        api_key (Optional[str]): Your RELAI API key. You can find this on RELAI Platform:
            Enterprise > Access Management > API Keys. API key can also be set
            as an environment variable `RELAI_API_KEY`.
        api_url (Optional[str]): The base URL for the RELAI API. Defaults to "https://api.relai.ai".
            API URL can also be set as an environment variable `RELAI_API_URL`.

    Raises:
        RELAIError: If the API key is not provided or cannot be found in the environment variables.
    """

    def __init__(self, api_key: Optional[str] = None, api_url: Optional[str] = None):
        super().__init__(api_key=api_key, api_url=api_url)

    @property
    def client(self) -> aiohttp.ClientSession:
        if self._client is None:
            self._client = aiohttp.ClientSession(base_url=self.api_url, headers=self.headers)
        return self._client

    async def close(self):
        if self._client is not None:
            await self._client.close()
            self._client = None

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.close()

    async def _request(self, method: str, url: str, **kwargs: Any) -> Any:
        """Performs an HTTP request and handles common errors."""
        try:
            async with self.client.request(method, url, **kwargs) as response:
                try:
                    response.raise_for_status()
                except aiohttp.ClientResponseError as e:
                    # Try to include response body in the error message
                    try:
                        error_text = await response.text()
                    except Exception:
                        error_text = "<failed to read response text>"
                    raise RELAIError(
                        f"HTTP error occurred: {e.status} - {e.message}\nResponse body: {error_text}"
                    ) from e
                return await response.json()

        except aiohttp.ClientError as e:
            raise RELAIError(f"Request failed due to {type(e).__name__}: {repr(e)}") from e

    async def _get(self, *args, **kwargs) -> Any:
        """Performs a GET request."""
        return await self._request("GET", *args, **kwargs)

    async def _post(self, *args, **kwargs) -> Any:
        """Performs a POST request."""
        return await self._request("POST", *args, **kwargs)

    async def _put(self, *args, **kwargs) -> Any:
        """Performs a PUT request."""
        return await self._request("PUT", *args, **kwargs)

    async def is_connected(self) -> bool:
        """Checks if the client can connect to the RELAI API."""

        try:
            await self._get("health/")
        except RELAIError:
            return False
        else:
            return True

    async def get_benchmark_keywords(self) -> list[str]:
        """
        Retrieves the list of benchmark keywords from the RELAI API.

        Returns:
            list[str]: A list of benchmark keywords.
        """
        response = await self._get("api/v1/benchmark/keywords/")
        return response

    async def get_benchmark(self, benchmark_id: str) -> list[dict[str, Any]]:
        """
        Retrieves the benchmark data for a given benchmark ID.

        Args:
            benchmark_id (str): The ID of the benchmark to retrieve.

        Returns:
            list[dict[str, Any]]: Samples from the benchmark.

        Raises:
            RELAIError: If the request fails or the benchmark is not found.
        """
        details_url = f"api/v1/benchmarks/all/{benchmark_id}/"

        metadata = await self._get(details_url)

        samples_data = await self._get(
            f"api/v1/benchmarks/all/{benchmark_id}/samples/", params={"no_pagination": "true"}
        )

        samples = []
        for s in samples_data:
            split = s.get("split", "All").lower().capitalize()
            sample = {
                "sample_uuid": s["sample_uuid"],
                "split": split,
                **metadata.get("hyperparameters", {}),  # Fields like `style_rubrics`, `format_rubrics`
                **s["column_values"],
            }
            samples.append(sample)
        return samples

    async def get_evaluator_response(
        self, evaluator_name: str, benchmark_id: str, sample_id: str, **kwargs: Any
    ) -> Any:
        """
        Calls a RELAI evaluator with the specified name and parameters.

        Args:
            evaluator_name (str): The name of the evaluator to call.
            **kwargs: Additional parameters to pass to the evaluator.

        Returns:
            Any: The response from the evaluator.

        Raises:
            RELAIError: If the request fails or the evaluator is not found.
        """
        payload = {"evaluator_type": evaluator_name, "benchmark_id": benchmark_id, "sample_id": sample_id, **kwargs}
        response = await self._post("api/v1/enterprise/evaluator/", json=payload)
        return response["evaluation_result"]["response"]

    async def _execute_maestro_task(self, url: str, data: dict, timeout: int = 1800) -> dict[str, Any]:
        """Executes a Maestro task by POSTing data and polling for the result."""
        response = await self._post(url, json=data)
        task_id = response["result"]["task_id"]
        return await self.poll_maestro_task(task_id, timeout=timeout)

    async def poll_maestro_task(self, task_id: str, timeout: int = 1800) -> Any:
        """
        Poll the status of a Maestro until it is complete or there is an error.
        If and when the task is complete, return the output.

        Args:
            task_id (str): The ID of the Maestro task to poll.
            timeout (int): The maximum time to wait for the task to complete (in seconds).

        Returns:
            Any: The output of the Maestro task if it is complete.

        Raises:
            RELAIError: If the polling request fails.
        """
        payload = {"task_id": task_id}
        sleep_interval = 0.25  # in seconds
        start = time.time()

        while True:
            elapsed = time.time() - start
            if elapsed > timeout:
                raise RELAIError(f"Maestro task timed out after {timeout:.2f} seconds")
            response = await self._post("api/v1/maestro/poll-task/", json=payload)
            status = response["result"]["status"]
            if status == "PENDING":
                await asyncio.sleep(sleep_interval)
                sleep_interval = min(sleep_interval * 2, 10)  # Exponential backoff, max 10 seconds
                continue
            elif status == "SUCCESS":
                return response["result"]
            else:
                raise RELAIError(f"Maestro task failed: {status}")

    async def process_test_case(self, data: dict) -> tuple[list[str], dict[str, int]]:
        """
        Process a new test case using the RELAI platform for later selecting samples for optimization.

        Args:
            data (dict): The data required for processing the test case.

        Returns:
            list[str]: A list of tags for the new test case.
            dict[str, int]: A dictionary of active tags and tag counts.

        Raises:
            RELAIError: If the processing request fails.
        """
        response = await self._execute_maestro_task("api/v1/maestro/process-test-case/", data)
        tags = response["response"]["tags"]
        updated_active_tags = response["response"]["updated_active_tags"]

        return tags, updated_active_tags

    async def apply_structure_to_code(self, data: dict) -> str:
        """
        Applies the suggested structural changes to the agent code with RELAI platform.

        Args:
            data (dict): The data required for applying structural changes to the agent code.

        Returns:
            str: The updated agent code after applying the suggested structural changes.

        Raises:
            RELAIError: If the application request fails.
        """
        response = await self._execute_maestro_task("api/v1/maestro/apply-structure-to-code/", data)

        return response["response"]["code"]

    async def optimize_structure(self, data: dict) -> str:
        """
        Optimizes the structure of an agent using the RELAI platform.

        Args:
            data (dict): The data required for structure optimization.

        Returns:
            str: The suggestion for optimizing the agent's structure.

        Raises:
            RELAIError: If the optimization request fails.
        """
        response = await self._execute_maestro_task("api/v1/maestro/optimize-structure/", data)
        return response["response"]["suggestion"]

    async def propose_values(self, data: dict) -> tuple[str, dict[str, str]]:
        """
        Proposes new values for the agent's parameters.

        Args:
            data (dict): The data required for proposing new values.

        Returns:
            tuple[str, dict[str, str]]: A tuple containing the analysis of agent implementation
                and the proposed parameter values.

        Raises:
            RELAIError: If the proposal request fails.
        """
        response = await self._execute_maestro_task("api/v1/maestro/propose-values/", data)
        analysis = response["response"]["analysis"]
        proposed_values = response["response"]["proposed_values"]
        return analysis, proposed_values

    async def review_values(self, data: dict) -> dict[str, Any]:
        """
        Reviews proposed values for the agent's parameters.

        Args:
            data (dict): The data required for reviewing proposed values.

        Returns:
            dict[str, Any]: The review of the proposed values.

        Raises::
            RELAIError: If the review request fails.
        """
        response = await self._execute_maestro_task("api/v1/maestro/review-values/", data)
        return response["response"]["review"]

    async def select_version(self, data: dict) -> int:
        """
        Selects the version to evaluate or apply

        Args:
            data (dict): The data to use for selecting the version.

        Returns:
            int: The ID of the selected version.

        Raises:
            RELAIError: If the selection request fails.
        """
        response = await self._execute_maestro_task("api/v1/maestro/select-version/", data)
        return response["response"]["version"]

    async def upload_trace(self, data: dict[str, Any]) -> str:
        """
        Uploads OpenInference execution logs of a run to the RELAI platform.

        Args:
            data (dict[str, Any]): A dictionary containing execution log data to upload.

        Returns:
            str: The id for the uploaded execution logs.

        Raises:
            RELAIError: If the upload request fails.
        """
        response = await self._post("/api/v1/traces/runs/", json=data)
        return response["trace_uuid"]

    async def upload_run(
        self,
        group_id: str,
        trace_id: str,
        agent_inputs: dict[str, Any] | None = None,
        agent_outputs: dict[str, Any] | None = None,
        extras: dict[str, Any] | None = None,
        serialized_simulation_config: dict[str, Any] | None = None,
        agent_version_uuid: str | None = None,
        environment_uuid: str | None = None,
    ) -> None:
        agent_inputs = agent_inputs or {}
        agent_outputs = agent_outputs or {}
        extras = extras or {}
        serialized_simulation_config = serialized_simulation_config or {}
        await self._post(
            "api/v1/agents/agent-runs/",
            json={
                "group_id": group_id,
                "trace": trace_id,
                "agent_inputs": agent_inputs,
                "agent_outputs": agent_outputs,
                "extras": extras,
                "serialized_simulation_config": serialized_simulation_config,
                "agent_version_uuid": agent_version_uuid,
                "environment_uuid": environment_uuid,
            },
        )

    async def upload_critico_log(
        self,
        trace_id: str,
        evaluator_logs: list[dict[str, Any]],
        aggregate_score: float,
        aggregate_feedback: str,
    ) -> None:
        """
        Submits evaluation data to the RELAI platform.

        Args:
            trace_id (str): A trace identifier for the corresponding agent simulation run.
            evaluator_logs (list[dict[str, Any]]): A list of logs from individual evaluators.
            aggregate_score (float): The aggregate score computed from all the evaluator logs.
            aggregate_feedback (str): The aggregate feedback compiled from all the evaluator logs.

        Returns:
            Any: The response from the submission.

        Raises:
            RELAIError: If the submission fails.
        """
        await self._post(
            "api/v1/agents/critico-logs/",
            json={
                "trace": trace_id,
                "evaluator_logs": evaluator_logs,
                "aggregate_score": aggregate_score,
                "aggregate_feedback": aggregate_feedback,
            },
        )

    async def upload_critico_report(self, title: str, raw_evaluation_sets: list[dict[str, Any]]) -> None:
        """
        Submit a Critico report to the RELAI platform.

        Args:
            title (str): The title of the Critico report.
            raw_evaluation_sets (list[dict[str, Any]]): The data of the Critico report.
        """
        await self._post(
            "/api/v1/benchmarks/evaluations/submit/",
            json={
                "title": title,
                "raw_evaluation_sets": raw_evaluation_sets,
            },
        )

    async def update_config_opt_visual(self, config_viz: ConfigOptVizSchema, uuid: str | None = None) -> str:
        """
        Updates the configuration optimization visualization data on the RELAI platform.

        Args:
            config_viz (ConfigOptVizSchema): The configuration optimization visualization data to update.
            uuid (str, optional): The UUID of the configuration optimization visualization to update. If not provided, a new visualization will be created.

        Returns:
            str: The uuid of the updated or created configuration optimization visualization.

        Raises:
            RELAIError: If the update request fails.
        """

        if uuid is None:
            response = await self._post("api/v1/maestro/config-optimizer/", json=config_viz.model_dump())
            return response["config_opt_viz_util_uuid"]
        else:
            await self._put(f"api/v1/maestro/config-optimizer/{uuid}/", json=config_viz.model_dump())
            return uuid

    async def update_graph_opt_visual(self, graph_vis: GraphOptVizSchema, uuid: str | None = None) -> str:
        """
        Updates the graph optimization visualization data on the RELAI platform.

        Args:
            graph_viz (GraphOptVizSchema): The graph optimization visualization data to update.
            uuid (str, optional): The UUID of the graph optimization visualization to update. If not provided, a new visualization will be created.

        Returns:
            str: The uuid of the updated or created graph optimization visualization.

        Raises:
            RELAIError: If the update request fails.
        """

        if uuid is None:
            response = await self._post("api/v1/maestro/graph-optimizer/", json=graph_vis.model_dump())
            return response["graph_opt_viz_util_uuid"]
        else:
            await self._put(f"api/v1/maestro/graph-optimizer/{uuid}/", json=graph_vis.model_dump())
            return uuid

    async def get_persona_set(self, persona_set_id: str) -> list[str]:
        """
        Retrieves the details of a persona set by its ID.

        Args:
            persona_set_id (str): The ID of the persona set to retrieve.

        Returns:
            list[str]: A list of system prompts in the persona set.
        """
        response = await self._get(f"api/v1/persona/persona-sets/{persona_set_id}/")
        return response["system_prompts"]

    async def get_custom_evaluator(self, evaluator_id: str) -> dict[str, Any]:
        """
        Retrieves the details of a custom evaluator by its ID.

        Args:
            evaluator_id (str): The ID of the evaluator to retrieve.

        Returns:
            dict[str, Any]: A dictionary containing the name and parameters of the evaluator.
        """
        response = await self._get(f"api/v1/benchmarks/functions/{evaluator_id}/")
        return {"name": response["name"], "parameters": response["parameters"]}


_default_client = RELAI()


def get_default_client() -> RELAI:
    return _default_client


def set_default_client(client: RELAI) -> None:
    global _default_client
    _default_client = client



from typing import Dict, List, Optional, Tuple, Union

from pydantic import BaseModel, Field


class ParamSchema(BaseModel):
    """
    Schema for an agent parameter.
    """

    type: str  # Type of the parameter
    current_value: Union[float, int, str]  # Current value of the parameter
    description: str  # Description of the parameter
    allowed_values: Optional[List[Union[float, int, str]]] = Field(default=None)  # Allowed values for the parameter


class RunSchema(BaseModel):
    """
    Schema for execution logs and evaluation results from a single run of an agent.
    """

    log: str  # Events in the Maestro execution log (a serialized json str)
    trace_id: str  # used to uniquely identify the openinference execution log
    input: str  # Input to the agent (serialized)
    output: str  # Output of the agent
    eval_score: float  # Evaluation score
    eval_feedback: str  # Evaluation feedback


class ConfigSchema(BaseModel):
    """
    Schema for a specific agent configuration and the corresponding statistics.
    """

    params: Dict[str, ParamSchema]  # Parameters. The keys are parameter names.
    validation_score: float  # Current average score on validation set
    validation_count: int  # Number of validation runs
    validation_runs: List[RunSchema]  # Execution logs & evaluation results (up to 100 runs)
    utc_timestamp: str  # Timestamp in UTC for when the configuration is saved (iso format)


class ConfigOptVizSchema(BaseModel):
    """
    Schema to be used for visualizing optimizer progress and results.
    """

    name: str
    configs: List[ConfigSchema]  # Different configurations saved during optimization and the corresponding statistics
    current_config: int  # Index of the current configuration
    validation_score_over_version: List[
        Tuple[int, float, int]
    ]  # Validation scores for different configs; Format: (config id, validation score, validation count)
    validation_score_over_time: List[
        Tuple[str, float, int]
    ]  # Validation scores over time; Format: (timestamp in UTC with iso format, validation score, validation count)


class GraphOptVizSchema(BaseModel):
    """
    Schema to be used for visualizing optimizer progress and results.
    """

    name: str
    proposal: str  # Proposed graph changes to the agent
    runs: List[RunSchema]  # Execution logs & evaluation results of the original agent, which motivate the proposal



import functools
import inspect
import random
import sys
import traceback
from collections.abc import Callable, Coroutine, Generator, Sequence
from contextlib import contextmanager
from contextvars import ContextVar
from typing import Any, ParamSpec, Protocol, TypeAlias, TypeVar, cast, get_type_hints
from uuid import uuid4

from relai.benchmark import Benchmark
from relai.mocker.base_mocker import BaseMocker

from ._client import RELAI, AsyncRELAI
from .data import AgentLog, AgentOutputs, RELAISample, Serializable, SimulationConfigT, SimulationTape
from .flags import tracking_off, tracking_on
from .utils import create_logging_span, get_current_logger

P = ParamSpec("P")
T = TypeVar("T")

SyncAgent: TypeAlias = Callable[[SimulationTape], AgentOutputs]
AsyncAgent: TypeAlias = Callable[[SimulationTape], Coroutine[None, None, AgentOutputs]]
Samplable: TypeAlias = Sequence[BaseMocker]


_simulation_mode: ContextVar[bool] = ContextVar("simulation_mode", default=False)
_simulation_config: ContextVar[SimulationConfigT] = ContextVar("simulation_config", default={})
_mocked_funcs: set[str] = set()


@contextmanager
def _simulate(config: SimulationConfigT):
    """
    A context manager to enter simulation mode with the given configuration.
    """
    simulation_mode_token = _simulation_mode.set(True)
    simulation_config_token = _simulation_config.set(config or {})
    yield
    _simulation_mode.reset(simulation_mode_token)
    _simulation_config.reset(simulation_config_token)


def get_current_simulation_mode() -> bool:
    """
    Retrieves the current simulation mode.

    Returns:
        bool: True if in simulation mode, False otherwise.
    """
    return _simulation_mode.get()


def get_current_simulation_config() -> SimulationConfigT:
    """
    Retrieves the current simulation configuration.

    Returns:
        dict[str, BaseMocker]: The current simulation configuration.
    """
    return _simulation_config.get()


def simulated(
    func: Callable[..., Coroutine[Any, Any, Any]] | Callable[P, T],
) -> Callable[..., Coroutine[Any, Any, Any]] | Callable[P, T]:
    """
    Decorator to mark a function to be simulated using a mocker in simulation mode. All such functions must have a
    corresponding mocker set in the simulation configuration. Supports both synchronous and asynchronous functions.
    """
    func_name = f"{func.__module__}.{func.__qualname__}"
    _mocked_funcs.add(func_name)

    func_doc = inspect.getdoc(func)
    mod_globals = sys.modules[func.__module__].__dict__
    try:
        type_hints = get_type_hints(func, globalns=mod_globals)
    except Exception:
        type_hints = dict(func.__annotations__)
    output_type = cast(type, type_hints.get("return", None))

    @functools.wraps(func)
    def sync_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        if not get_current_simulation_mode():
            return cast(Callable[P, T], func)(*args, **kwargs)
        else:
            try:
                mocker = get_current_simulation_config()[func_name]
            except KeyError as e:
                raise RuntimeError(
                    f"No mocker found for function {func_name}. Make sure to set the mocker in the simulation config."
                ) from e
            if mocker.func_doc is None:
                mocker.func_doc = func_doc
            if mocker.output_type is None:
                mocker.output_type = output_type
            return mocker._run(*args, **kwargs)

    @functools.wraps(func)
    async def async_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        if not get_current_simulation_mode():
            return await cast(Callable[P, Coroutine[Any, Any, T]], func)(*args, **kwargs)
        else:
            try:
                mocker = get_current_simulation_config()[func_name]
            except KeyError as e:
                raise RuntimeError(
                    f"No mocker found for function {func_name}. Make sure to set the mocker in the simulation config."
                ) from e
            if mocker.func_doc is None:
                mocker.func_doc = func_doc
            if mocker.output_type is None:
                mocker.output_type = output_type
            return await mocker._arun(*args, **kwargs)

    if inspect.iscoroutinefunction(func):
        return async_wrapper
    else:
        return sync_wrapper


class EnvGenerator(Protocol):
    """
    Protocol for functions that generate simulation configurations. These functions take an optional RELAISample
    and return a SimulationConfigT which is a dictionary mapping qualified function names of functions decorated with
    `@simulated` to their respective mocker instances.
    """

    def __call__(self, sample: RELAISample | None = None) -> SimulationConfigT: ...


def _default_env_generator(sample: RELAISample | None = None) -> SimulationConfigT:
    return {}


def random_env_generator(config_set: dict[str, Samplable]) -> EnvGenerator:
    """
    An environment generator that uniformly samples a mocker for each simulated function from the provided set of
    mockers.

    Args:
        config_set (dict[str, Sequence[BaseMocker]]): A mapping from qualified function names to a sequence of possible
            mockers for that function.
    """

    def sampler(sample: RELAISample | None = None) -> SimulationConfigT:
        return {k: random.choice(v) for k, v in config_set.items()}

    return sampler


class BaseSimulator:
    def __init__(self, env_generator: EnvGenerator | None = None, benchmark: Benchmark | None = None) -> None:
        if env_generator is None:
            self.env_generator = _default_env_generator
        else:
            self.env_generator = env_generator
        self.benchmark = benchmark

    def tape_and_config_generator(
        self, num_runs: int
    ) -> Generator[tuple[SimulationTape, SimulationConfigT], None, None]:
        if self.benchmark is None:
            for _ in range(num_runs):
                config = self.env_generator()
                tape = SimulationTape()
                tape.simulation_config = config
                yield tape, config
        else:
            samples = self.benchmark.sample(num_runs)
            for sample in samples:
                config = self.env_generator(sample)
                tape = SimulationTape(sample=sample)
                tape.simulation_config = config
                yield tape, config


class SyncSimulator(BaseSimulator):
    """A simulator for synchronous agent functions."""

    def __init__(
        self,
        agent_fn: SyncAgent,
        env_generator: EnvGenerator | None = None,
        benchmark: Benchmark | None = None,
        log_runs: bool = True,
        client: RELAI | None = None,
    ) -> None:
        """
        Args:
            agent_fn (SyncAgent): The synchronous agent function to be simulated.
            env_generator (EnvGenerator | None): An optional environment generator function. If not provided, a default
                generator that returns an empty configuration will be used. The default generator can only be used with
                simulations that don't require any mockers.
            benchmark (Benchmark | None): An optional benchmark to source simulation samples from.
            log_runs (bool): Whether to log the runs to the RELAI platform. Defaults to True.
            client (RELAI | None): A synchronous RELAI client for logging. Must be provided if log_runs is True.
        """
        super().__init__(env_generator, benchmark)
        self.agent_fn = agent_fn
        self.log_runs = log_runs
        if log_runs and client is None:
            raise ValueError("client must be provided if log_runs is True")
        self.client = client

    def run(
        self,
        num_runs: int,
        group_id: str | None = None,
        agent_version_uuid: str | None = None,
        environment_uuid: str | None = None,
        catch_exception: bool = False,
    ) -> list[AgentLog]:
        """
        Run the simulator for a specified number of times.

        Args:
            num_runs (int): The number of simulation runs to execute.
            group_id (str, optional): An optional group ID to associate all runs together. If not provided,
                a new UUID will be generated.
            agent_version_uuid (str, optional): An optional agent version UUID to associate with the runs.
            environment_uuid (str, optional): An optional environment UUID to associate with the runs.
        """
        agent_logs: list[AgentLog] = []
        group_id = ("Simulate-" + uuid4().hex) if group_id is None else group_id
        tracking_on()
        for tape, config in self.tape_and_config_generator(num_runs):
            with _simulate(config), create_logging_span(tape.id):
                if catch_exception:
                    try:
                        agent_outputs = self.agent_fn(tape)
                    except Exception:
                        agent_outputs = {"[Exception]": traceback.format_exc(limit=3)}
                else:
                    agent_outputs = self.agent_fn(tape)

                if not isinstance(agent_outputs, dict):
                    raise TypeError("agent_fn must return an instance of AgentOutputs")
                agent_log = AgentLog(simulation_tape=tape, agent_outputs=agent_outputs)
                if self.log_runs and self.client is not None:
                    trace_id = self.client.upload_trace(data=get_current_logger().to_openinference())
                    agent_log.trace_id = trace_id
                    self.client.upload_run(
                        group_id=group_id,
                        trace_id=trace_id,
                        agent_inputs=tape.agent_inputs,
                        agent_outputs=agent_outputs,
                        extras=tape.extras,
                        serialized_simulation_config={
                            k: v.serialize() for k, v in tape.simulation_config.items() if isinstance(v, Serializable)
                        },
                        agent_version_uuid=agent_version_uuid,
                        environment_uuid=environment_uuid,
                    )
                agent_logs.append(agent_log)
                tape.add_record("relai_log", get_current_logger().serialize())
        tracking_off()
        return agent_logs

    def rerun(
        self,
        simulation_tapes: list[SimulationTape],
        group_id: str | None = None,
        agent_version_uuid: str | None = None,
        environment_uuid: str | None = None,
        catch_exception: bool = False,
    ) -> list[AgentLog]:
        """
        Rerun the simulator for a list of simulation tapes.

        Args:
            simulation_tapes (list[SimulationTape]): The list of simulation tapes to rerun. This allows for re-executing
                the agent in an environment identical to a previous run and is useful for debugging and optimization.
            group_id (str, optional): An optional group ID to associate all runs together. If not provided,
                a new UUID will be generated.
            agent_version_uuid (str, optional): An optional agent version UUID to associate with the runs.
            environment_uuid (str, optional): An optional environment UUID to associate with the runs.
            catch_exception (bool, optional): Whether to catch exceptions during simulation runs. Defaults to False.
        """
        agent_logs: list[AgentLog] = []
        group_id = ("Simulate-" + uuid4().hex) if group_id is None else group_id
        tracking_on()
        for tape in simulation_tapes:
            new_tape = tape.copy()
            with _simulate(new_tape.simulation_config), create_logging_span(new_tape.id):
                if catch_exception:
                    try:
                        agent_outputs = self.agent_fn(new_tape)
                    except Exception:
                        agent_outputs = {"[Exception]": traceback.format_exc(limit=3)}
                else:
                    agent_outputs = self.agent_fn(new_tape)

                agent_log = AgentLog(simulation_tape=new_tape, agent_outputs=agent_outputs)
                agent_logs.append(agent_log)
                if self.log_runs and self.client is not None:
                    trace_id = self.client.upload_trace(data=get_current_logger().to_openinference())
                    self.client.upload_run(
                        group_id=group_id,
                        trace_id=trace_id,
                        agent_inputs=new_tape.agent_inputs,
                        agent_outputs=agent_outputs,
                        extras=new_tape.extras,
                        serialized_simulation_config={
                            k: v.serialize()
                            for k, v in new_tape.simulation_config.items()
                            if isinstance(v, Serializable)
                        },
                        agent_version_uuid=agent_version_uuid,
                        environment_uuid=environment_uuid,
                    )
                new_tape.add_record("relai_log", get_current_logger().serialize())
        tracking_off()
        return agent_logs


class AsyncSimulator(BaseSimulator):
    """A simulator for asynchronous agent functions."""

    def __init__(
        self,
        agent_fn: AsyncAgent,
        env_generator: EnvGenerator | None = None,
        benchmark: Benchmark | None = None,
        log_runs: bool = True,
        client: AsyncRELAI | None = None,
    ) -> None:
        """
        Args:
            agent_fn (AsyncAgent): The asynchronous agent function to be simulated.
            env_generator (EnvGenerator | None): An optional environment generator function. If not provided, a default
                generator that returns an empty configuration will be used. The default generator can only be used with
                simulations that don't require any mockers.
            benchmark (Benchmark | None): An optional benchmark to source simulation samples from.
            log_runs (bool): Whether to log the runs to the RELAI platform. Defaults to True.
            client (RELAI | None): A asynchronous RELAI client for logging. Must be provided if log_runs is True.
        """
        super().__init__(env_generator, benchmark)
        self.agent_fn = agent_fn
        self.log_runs = log_runs
        if log_runs and client is None:
            raise ValueError("client must be provided if log_runs is True")
        self.client = client

    async def run(
        self,
        num_runs: int,
        group_id: str | None = None,
        agent_version_uuid: str | None = None,
        environment_uuid: str | None = None,
        catch_exception: bool = False,
    ) -> list[AgentLog]:
        """Run the simulator for a specified number of times.

        Args:
            num_runs (int): The number of simulation runs to execute.
            group_id (str, optional): An optional group ID to associate all runs together. If not provided,
                a new UUID will be generated.
            agent_version_uuid (str, optional): An optional agent version UUID to associate with the runs.
            environment_uuid (str, optional): An optional environment UUID to associate with the runs.
            catch_exception (bool): Whether to catch errors during simulation runs. Defaults to False.
        """
        agent_logs: list[AgentLog] = []
        group_id = ("Simulate-" + uuid4().hex) if group_id is None else group_id
        tracking_on()
        for tape, config in self.tape_and_config_generator(num_runs):
            with _simulate(config), create_logging_span(tape.id):
                if catch_exception:
                    try:
                        agent_outputs = await self.agent_fn(tape)
                    except Exception:
                        agent_outputs = {"[Exception]": traceback.format_exc(limit=3)}
                else:
                    agent_outputs = await self.agent_fn(tape)

                agent_log = AgentLog(simulation_tape=tape, agent_outputs=agent_outputs)
                if self.log_runs and self.client is not None:
                    trace_id = await self.client.upload_trace(data=get_current_logger().to_openinference())
                    agent_log.trace_id = trace_id
                    await self.client.upload_run(
                        group_id=group_id,
                        trace_id=trace_id,
                        agent_inputs=tape.agent_inputs,
                        agent_outputs=agent_outputs,
                        extras=tape.extras,
                        serialized_simulation_config={
                            k: v.serialize() for k, v in tape.simulation_config.items() if isinstance(v, Serializable)
                        },
                        agent_version_uuid=agent_version_uuid,
                        environment_uuid=environment_uuid,
                    )
                agent_logs.append(agent_log)
                tape.add_record("relai_log", get_current_logger().serialize())
        tracking_off()
        return agent_logs

    async def rerun(
        self,
        simulation_tapes: list[SimulationTape],
        group_id: str | None = None,
        agent_version_uuid: str | None = None,
        environment_uuid: str | None = None,
        catch_exception: bool = False,
    ) -> list[AgentLog]:
        """
        Rerun the simulator for a list of simulation tapes.

        Args:
            simulation_tapes (list[SimulationTape]): The list of simulation tapes to rerun. This allows for re-executing
                the agent in an environment identical to a previous run and is useful for debugging and optimization.
            group_id (str, optional): An optional group ID to associate all runs together. If not provided,
                a new UUID will be generated.
            agent_version_uuid (str, optional): An optional agent version UUID to associate with the runs.
            environment_uuid (str, optional): An optional environment UUID to associate with the runs.
            catch_exception (bool): Whether to catch exceptions during simulation runs. Defaults to False.
        """
        agent_logs: list[AgentLog] = []
        group_id = ("Simulate-" + uuid4().hex) if group_id is None else group_id
        tracking_on()
        for tape in simulation_tapes:
            new_tape = tape.copy()
            with _simulate(new_tape.simulation_config), create_logging_span(new_tape.id):
                if catch_exception:
                    try:
                        agent_outputs = await self.agent_fn(new_tape)
                    except Exception:
                        agent_outputs = {"[Exception]": traceback.format_exc(limit=3)}
                else:
                    agent_outputs = await self.agent_fn(new_tape)

                agent_log = AgentLog(simulation_tape=new_tape, agent_outputs=agent_outputs)
                agent_logs.append(agent_log)
                if self.log_runs and self.client is not None:
                    trace_id = await self.client.upload_trace(data=get_current_logger().to_openinference())
                    await self.client.upload_run(
                        group_id=group_id,
                        trace_id=trace_id,
                        agent_inputs=new_tape.agent_inputs,
                        agent_outputs=agent_outputs,
                        extras=new_tape.extras,
                        serialized_simulation_config={
                            k: v.serialize()
                            for k, v in new_tape.simulation_config.items()
                            if isinstance(v, Serializable)
                        },
                        agent_version_uuid=agent_version_uuid,
                        environment_uuid=environment_uuid,
                    )
                new_tape.add_record("relai_log", get_current_logger().serialize())
        tracking_off()
        return agent_logs



from functools import cached_property
from typing import ClassVar
from uuid import uuid4

from agents import Agent, Runner, SQLiteSession

from ..utils import no_trace
from .base_mocker import BaseMocker


class MockTool(BaseMocker):
    """A mocker class for simulating the behavior of a tool used by an AI agent."""

    prompt_template: ClassVar[str] = (
        """You are an assistant helping with testing another AI agent. Your job is to """
        """mock the output for a tool based on the description of the tool (provided below) """
        """and the additional context (if any). Only respond """
        """with what the tool might output, without any extraneous phrases like "Here's the output".\n"""
        """<description>{description}</description>\n"""
        """<additional context>{context}</additional context>\n"""
    )

    def __init__(
        self,
        model: str | None = "gpt-5-mini",
        context: str | None = None,
    ):
        """
        Initializes the MockTool with an optional model specification.

        Args:
            model (str | None): The AI model to use for simulating the tool's behavior.
            context (str | None): Additional context to guide behavior of the mock tool.
        """
        super().__init__()
        self.name = f"mock-tool-{uuid4().hex}"
        self.model = model
        self.context = context
        self._session = SQLiteSession(self.name)

    @cached_property
    def agent(self) -> Agent:
        return Agent(
            name=self.name,
            instructions=self.prompt_template.format(description=self._func_doc, context=self.context or ""),
            model=self.model,
            output_type=self.output_type,
        )

    def _run(self, *args, **kwargs):
        agent_input = str(
            {
                "args": args,
                "kwargs": kwargs,
            }
        )
        with no_trace():
            result = Runner.run_sync(
                self.agent,
                agent_input,
                session=self._session,
            )
        output = result.final_output
        return output

    async def _arun(self, *args, **kwargs):
        agent_input = str(
            {
                "args": args,
                "kwargs": kwargs,
            }
        )
        with no_trace():
            result = await Runner.run(
                self.agent,
                agent_input,
                session=self._session,
            )
        output = result.final_output
        return output

    def serialize(self) -> dict[str, str]:
        return {
            "model": self.model or "None",
            "context": self.context or "None",
        }



from .optimizer import Maestro
from .params import RegisteredParameters, get_current_params, params, register_param, set_current_params
from .utils import component

__all__ = [
    "Maestro",
    "RegisteredParameters",
    "get_current_params",
    "params",
    "register_param",
    "set_current_params",
    "component",
]



import asyncio
import copy
import json
import os
from datetime import datetime, timezone
from typing import Any, Awaitable, Callable, Optional
from uuid import uuid4

from tqdm.auto import tqdm

from relai import AsyncRELAI
from relai.critico.critico import Critico, CriticoLog
from relai.simulator import AgentLog, AsyncAgent, AsyncSimulator
from relai.utils import no_trace

from ..schema.visual import ConfigOptVizSchema, ConfigSchema, GraphOptVizSchema, ParamSchema, RunSchema
from .graph import get_current_param_graph
from .params import get_current_params
from .utils import ProportionalSampler, extract_code, get_full_func_name


class Maestro:
    """
    Maestro automatically optimizes an AI agent to maximize its Critico score, navigating the space of
    configurations to intelligently improve performance on the chosen criteria.
    """

    def __init__(
        self,
        client: AsyncRELAI,
        agent_fn: AsyncAgent,
        goal: Optional[str] = None,
        max_memory: int = 20,
        name: str = "No Name",
        log_to_platform: bool = True,
    ):
        """
        Args:
            client (AsyncRELAI): An instance of the AsyncRELAI client to interact with the RELAI platform.
            agent_fn (AsyncAgent): The agent function to be optimized.
            goal (str, optional): Optional description of the goal of optimization. If None, increasing evaluation score
                will be considered as the only goal. Defaults to None.
            max_memory (int, optional): Control the maximum number of previous optimization history visible at each
                optimization step. Defaults to 20.
            name (str, optional): Name of the configuration optimization visualization on RELAI platform.
                Defaults to "No Name".
            log_to_platform (bool): Whether to log optimization progress and results on RELAI platform.
                Defaults to True.
        """
        self.setups: list[dict] = []
        self.agent_fn: AsyncAgent = agent_fn
        self.log: list[dict] = []
        self.max_memory: int = max_memory
        self._client: AsyncRELAI = client
        self.goal: str = goal if goal is not None else "Higher scores"
        self.log_to_platform: bool = log_to_platform
        self.config_opt_viz_id: str | None = None
        self.name: str = name

        self.versions: list[dict] = [
            {
                "idx": 0,
                "params": copy.deepcopy(get_current_params()),
                "log": copy.deepcopy(self.log),
                "visits": 0,
                "average_score": 0.0,
                "utc_timestamp": datetime.now(timezone.utc).isoformat(),
                "validation_runs": [],
                "validation_count": 0,
            }
        ]
        self.current_version: int = 0
        self.total_visits: int = 0
        self.inf: float = 1e208
        self.score_min: float = self.inf
        self.score_max: float = -self.inf

    def add_setup(self, simulator: AsyncSimulator, critico: Critico, weight: float = 1):
        """
        Add a new setup consisting of a simulator and a critico to Maestro.

        Args:
            simulator (AsyncSimulator): An AsyncSimulator to run the agent in the new setup.
            critico (Critico): A Critico with evaluators for the new setup.
            weight (float): A positive float representing the weight of this setup in comparson to others.
                Defaults to 1.
        """

        if not isinstance(simulator, AsyncSimulator):
            raise ValueError("Currently only AsyncSimulator is supported in Maestro.")
        if weight <= 0:
            raise ValueError(f"`weight` must be a positive float, got {weight}.")

        self.setups.append({"simulator": simulator, "critico": critico, "weight": weight})

    def _serialize_past_proposals(self) -> str:
        """
        Serialize the previous optimization history of the current version.

        Returns:
            str: A string representing the previous optimization history of the current version.
        """
        return json.dumps(self.log[-self.max_memory :], indent=2)

    async def _select(self, explore: bool = False):
        """
        Select the best performing or the most promising version (of the agent to optimize)
        according to past evidences and set it as the current version.

        Args:
            explore (bool): If False, setting the best performing version as current;
                If True, setting the most promising version as current. Defaults to False.
        """
        selected_version = await self._client.select_version(
            {
                "versions": [{"average_score": v["average_score"], "visits": v["visits"]} for v in self.versions],
                "explore": explore,
                "min_score": self.score_min,
                "max_score": self.score_max,
                "total_visits": self.total_visits,
            }
        )
        self.current_version = selected_version
        get_current_params().sync(self.versions[selected_version]["params"])
        self.log = copy.deepcopy(self.versions[selected_version]["log"])

    def _update_score(self, score: float):
        """
        Adding an evaluation score of the current version to evidence for version selection.

        Args:
            score (float): The evaluation score of the current version to be added as evidence
            for version selection.
        """
        self.total_visits += 1
        self.versions[self.current_version]["average_score"] = (
            self.versions[self.current_version]["average_score"] * self.versions[self.current_version]["visits"]
            + score
        ) / (self.versions[self.current_version]["visits"] + 1.0)
        self.versions[self.current_version]["visits"] += 1

        self.score_min = min(self.score_min, score)
        self.score_max = max(self.score_max, score)

    def _serialize_agent_outputs(self, agent_outputs: Any) -> str:
        """
        Serialize agent outputs.

        Args:
            agent_outputs (Any): The agent outputs to serialize.

        Returns:
            str: A string representing the agent outputs.
        """
        if isinstance(agent_outputs, dict):
            ret = ""
            for key, value in agent_outputs.items():
                ret += f"{key}:\n {value}\n"
            return ret
        else:
            return str(agent_outputs)

    async def _evaluate(
        self, awaitables: list[Awaitable], criticos: list[Critico], verbose: bool = True, print_flag: str = ""
    ) -> tuple[list[dict[str, Any]], list[AgentLog]]:
        """
        Run and evaluate the current version of the agent through a set of awaitables.

        Args:
            awaitables (list[Awaitable]): A list of awaitables, each representing a run of the agent
            criticos (list[Critico]): A list of Critico objects, each corresponding to an awaitable
            verbose (bool): If True, additional information will be printed during evaluation.
                Defaults to True.
            print_flag (str): A string to be put next to the printed info when `verbose` is True.
                Used to distinguish printed info from different types of evaluations.

        Returns:
            list[dict]: A list of dictionary. Each dictionary contains evaluation info corresponding to
                one `Sample` object and has the following keys: "input", "log", "output", "eval_score",
                "eval_feedback".
        """
        RELAI_EVAL_BATCH_SIZE = int(os.getenv("RELAI_EVAL_BATCH_SIZE", 10))

        get_current_param_graph().clear()
        test_cases = []

        async def _evaluate_awaitable(awaitable: Awaitable, critico: Critico) -> tuple[dict[str, Any], AgentLog]:
            agent_log: AgentLog = (await awaitable)[0]

            # Handle errors gracefully
            if agent_log.agent_outputs.get("[Exception]") is not None:
                return {
                    "input": str(agent_log.simulation_tape.agent_inputs),
                    "log": agent_log.simulation_tape.extras["relai_log"],
                    "trace_id": agent_log.trace_id,
                    "output": self._serialize_agent_outputs(agent_log.agent_outputs),
                    "eval_score": 0.0,
                    "eval_feedback": (
                        "The following error is encountered during agent execution and should be fixed:\n"
                        f"{agent_log.agent_outputs.get('[Exception]')}"
                    ),
                }, agent_log

            with no_trace():
                eval_result: CriticoLog = (await critico.evaluate([agent_log]))[0]

            return {
                "input": str(agent_log.simulation_tape.agent_inputs),
                "log": agent_log.simulation_tape.extras["relai_log"],
                "trace_id": agent_log.trace_id,
                "output": self._serialize_agent_outputs(agent_log.agent_outputs),
                "eval_score": eval_result.aggregate_score,
                "eval_feedback": eval_result.aggregate_feedback,
            }, agent_log

        batches_awaitable = [
            awaitables[i : i + RELAI_EVAL_BATCH_SIZE] for i in range(0, len(awaitables), RELAI_EVAL_BATCH_SIZE)
        ]
        batches_critico = [
            criticos[i : i + RELAI_EVAL_BATCH_SIZE] for i in range(0, len(criticos), RELAI_EVAL_BATCH_SIZE)
        ]
        test_cases = []
        agent_logs = []
        for batch_awaitable, batch_critico in zip(batches_awaitable, batches_critico):
            tasks = [_evaluate_awaitable(sample, critico) for sample, critico in zip(batch_awaitable, batch_critico)]
            batch_test_cases = await asyncio.gather(*tasks)
            test_cases.extend([test_case for test_case, _ in batch_test_cases])
            agent_logs.extend([agent_log for _, agent_log in batch_test_cases])

        if verbose:
            for test_case in test_cases:
                print("=================agent excution result===================")
                print(f"- input:\n{test_case['input']}\n")
                print(f"- log{print_flag}:\n{test_case['log']}\n")
                print(f"- output{print_flag}:\n{test_case['output']}\n")
                print(f"- eval score{print_flag}:\n{test_case['eval_score']}\n")
                print(f"- eval feedback{print_flag}:\n{test_case['eval_feedback']}\n")
                print("=========================================================\n\n")

        return test_cases, agent_logs

    async def _iterate(
        self,
        batch_size: int,
        sampler: ProportionalSampler,
        verbose: bool = True,
        group_id: str | None = None,
        pbar: tqdm | None = None,
        agent_version_uuid: str | None = None,
        environment_uuid: str | None = None,
    ) -> bool:
        """
        An iterate step will propose changes to the current version of the agent and
        conduct a preliminary examination of the proposed changes.

        It returns True if the proposed changes pass the preliminary examination and
        False otherwise.

        Args:
            batch_size (int): 2 * `batch_size` samples will be obtained from the main Critico agent,
                i.e. `critico`, where `batch_size` of them will be used to propose changes and the other
                `batch_size` of them will be used for preliminary examinations.
            sampler (ProportionalSampler): Sampler to use for selecting setups.
            verbose (bool): If True, additional information will be printed during the iterate step.
                Defaults to True.
            group_id (str, optional): An optional group ID to associate all runs together. If not provided,
                a new UUID will be generated.
            pbar (tqdm, optional): A progress bar to display the progress of the iteration. Defaults to None.
            agent_version_uuid (str, optional): An optional agent version UUID to associate with the runs.
            environment_uuid (str, optional): An optional environment UUID to associate with the runs.

        Returns:
            bool: True if the proposed changes pass the preliminary examination and False otherwise.

        Raises:
            ValueError: If no setup (simulator, critico) has been added to Maestro.
        """
        if len(self.setups) == 0:
            raise ValueError(
                "No setup (simulator, critico) has been added to Maestro. Please add at least one setup before optimization."
            )

        group_id = uuid4().hex if group_id is None else group_id

        setups = sampler.sample(batch_size * 2)
        awaitables = []
        criticos = []
        for setup in setups:
            simulator = setup["simulator"]
            critico = setup["critico"]
            awaitables.append(
                simulator.run(
                    num_runs=1,
                    group_id=group_id,
                    agent_version_uuid=agent_version_uuid,
                    environment_uuid=environment_uuid,
                    catch_exception=True,
                )
            )
            criticos.append(critico)

        test_cases, agent_logs = await self._evaluate(awaitables=awaitables, criticos=criticos, verbose=verbose)

        if pbar is not None:
            pbar.update(len(test_cases))

        analysis, proposed_values = await self._client.propose_values(
            {
                "params": get_current_params().export(),
                "serialized_past_proposals": self._serialize_past_proposals(),
                "test_cases": test_cases[:batch_size],
                "goal": self.goal,
                "param_graph": get_current_param_graph().export(),
            }
        )

        changes = []
        for param, value in proposed_values.items():
            changes.append(
                {"param": param, "previous value": get_current_params().__getattr__(param), "new value": value}
            )
            if verbose:
                print("=" * 60)
                print("- proposed param change:", param)
                print("")
                print("- previous value:\n\n", get_current_params().__getattr__(param))
                print("")
                print("- new value:\n\n", value)
                print("=" * 60)

        self.log.append({"proposal id": len(self.log), "proposed changes": changes})

        # Decide if the change should be kept

        for name, proposed_value in proposed_values.items():
            get_current_params().update(name, proposed_value)

        new_awaitables = []
        new_criticos = []
        for test_case, agent_log, setup in zip(test_cases, agent_logs, setups):
            simulator = setup["simulator"]
            critico = setup["critico"]
            new_awaitables.append(
                simulator.rerun(
                    [agent_log.simulation_tape],
                    group_id=group_id,
                    agent_version_uuid=agent_version_uuid,
                    environment_uuid=environment_uuid,
                    catch_exception=True,
                )
            )
            new_criticos.append(critico)

        test_cases_updated, _ = await self._evaluate(
            awaitables=new_awaitables, criticos=new_criticos, verbose=verbose, print_flag=" (changed)"
        )

        if pbar is not None:
            pbar.update(len(test_cases_updated))

        for sample_id in range(0, batch_size * 2):
            test_cases_updated[sample_id]["previous_log"] = test_cases[sample_id]["log"]
            test_cases_updated[sample_id]["previous_output"] = test_cases[sample_id]["output"]
            test_cases_updated[sample_id]["previous_eval_score"] = test_cases[sample_id]["eval_score"]
            test_cases_updated[sample_id]["previous_eval_feedback"] = test_cases[sample_id]["eval_feedback"]

        for change in changes:
            get_current_params().update(change["param"], change["previous value"])

        previous_score = 0
        new_score = 0
        for test_case in test_cases_updated[batch_size:]:
            previous_score += test_case["previous_eval_score"]
            new_score += test_case["eval_score"]
        previous_score = float(previous_score) / batch_size
        new_score = float(new_score) / batch_size

        review_decision = await self._client.review_values(
            {
                "params": get_current_params().export(),
                "serialized_past_proposals": self._serialize_past_proposals(),
                "proposal": changes,
                "test_cases": test_cases_updated[:batch_size],
                "holdout_test_cases": test_cases_updated[batch_size:],
                "previous_score": previous_score,
                "new_score": new_score,
                "goal": self.goal,
                "analysis": analysis,
            }
        )

        if review_decision["accepted"]:
            self.log[-1]["status"] = "ACCEPTED"
            self.log[-1]["review comment"] = review_decision["full comment"]
            for change in changes:
                get_current_params().update(change["param"], change["new value"])

        else:
            self.log[-1]["status"] = "REJECTED"
            self.log[-1]["review comment"] = review_decision["full comment"]

        if verbose:
            print("previous avg score: ", previous_score)
            print("new avg score: ", new_score)
            print("accepted: ", review_decision["accepted"])
            print("review comment:\n", review_decision["full comment"])
            print("-" * 60 + "\n\n")

        return review_decision["accepted"]

    async def optimize_config(
        self,
        total_rollouts: int,
        batch_size: int = 8,
        explore_radius: int = 5,
        explore_factor: float = 0.5,
        group_id: str | None = None,
        verbose: bool = True,
        agent_version_uuid: str | None = None,
        environment_uuid: str | None = None,
    ) -> dict[str, Any]:
        """
        Optimize the configs (parameters) of the agent.

        Args:
            total_rollouts (int): Total number of rollouts to use for optimization.
            batch_size (int): Base batch size to use for individual optimization steps. Defaults to 8.
            explore_radius (int): A positive integer controlling the aggressiveness of exploration during optimization.
                A larger `explore_radius` encourages the optimizer to make more substantial changes between successive configurations.
                Defaults to 5.
            explore_factor (float): A float between 0 to 1 controlling the exploration-exploitation trade-off.
                A higher `explore_factor` allocates more rollouts to discover new configs,
                while a lower value allocates more rollouts to ensure the discovered configs are thoroughly evaluated.
                Defaults to 0.5.
            group_id (str, optional): An optional group ID to associate all runs together. If not provided,
                a new UUID will be generated.
            verbose (bool): If True, related information will be printed during the optimization step.
                Defaults to True.
            agent_version_uuid (str, optional): An optional agent version UUID to associate with the runs.
            environment_uuid (str, optional): An optional environment UUID to associate with the runs.

        Raises:
            ValueError: If the input parameters are not valid.

        Returns:
            dict[str, Any]: Dictionary representation of all optimized parameters.
        """
        if not isinstance(batch_size, int) or batch_size <= 0:
            raise ValueError(f"`batch_size` must be a positive integer, got {batch_size}.")

        if not isinstance(explore_radius, int) or explore_radius <= 0:
            raise ValueError(f"`explore_radius` must be a positive integer, got {explore_radius}.")

        if explore_factor <= 0 or explore_factor >= 1:
            raise ValueError(f"`explore_factor` must be a float between 0 and 1, got {explore_factor}.")

        group_size = (batch_size + 1) // 2
        # total_rollouts = (iterate_steps * group_size * 4 + select_steps * group_size) * num_rounds
        # explore_factor = (iterate_steps * group_size * 4) / (iterate_steps * group_size * 4 + select_steps * group_size)
        iterate_steps: int = explore_radius
        select_steps: int = int(explore_radius * 4 * (1 - explore_factor) / explore_factor)
        num_rounds: int = int(total_rollouts / (iterate_steps * group_size * 4 + select_steps * group_size))
        total_rollouts = num_rounds * (iterate_steps * group_size * 4 + select_steps * group_size)

        print("\n" + "=" * 30 + "optimize_config settings" + "=" * 30)
        print("  total_rollouts: ", total_rollouts)
        print("  (adjusted) batch_size: ", group_size * 2)
        print("  explore_radius: ", explore_radius)
        print("  explore_factor: ", explore_factor)
        print("-" * 60)
        print("  num_rounds: ", num_rounds)
        print("  steps per round for config discovery: ", iterate_steps)
        print("  steps per round for config evaluation: ", select_steps)

        print("=" * 80 + "\n\n")

        if num_rounds == 0:
            raise ValueError(
                f"`total_rollouts` is too small for the given `batch_size` {batch_size}, `explore_radius` {explore_radius}, and `explore_factor` {explore_factor}. "
                f"Please increase `total_rollouts` to at least {iterate_steps * group_size * 4 + select_steps * group_size}."
            )

        sampler = ProportionalSampler(
            elements=self.setups,
            weights=[setup["weight"] for setup in self.setups],
        )
        if group_id is None:
            group_id = "Maestro-Config-" + uuid4().hex
        pbar = tqdm(total=total_rollouts, desc="Total rollouts consumed for config optimization")

        for round in range(num_rounds):
            print("\n\n" + "=" * 30 + f" Round {round + 1}/{num_rounds} begins" + "=" * 30)
            print("Total versions accepted: ", len(self.versions))
            print("Rebase to version: ", self.current_version)
            print(
                "Score for the current base version: %s based on %s rollouts"
                % (
                    self.versions[self.current_version]["average_score"],
                    self.versions[self.current_version]["visits"] * group_size,
                )
            )
            print("\n\n")

            new_version = False
            for _ in range(iterate_steps):
                changes_accepted = await self._iterate(
                    batch_size=group_size,
                    verbose=verbose,
                    sampler=sampler,
                    group_id=group_id,
                    pbar=pbar,
                    agent_version_uuid=agent_version_uuid,
                    environment_uuid=environment_uuid,
                )
                if changes_accepted:
                    new_version = True

            if new_version:
                self.current_version = len(self.versions)
                self.versions.append(
                    {
                        "idx": self.current_version,
                        "params": copy.deepcopy(get_current_params()),
                        "log": copy.deepcopy(self.log[-self.max_memory :]),
                        "visits": 0,
                        "average_score": 0.0,
                        "utc_timestamp": datetime.now(timezone.utc).isoformat(),
                        "validation_runs": [],
                        "validation_count": 0,
                    }
                )

            # Evaluate existing versions
            for _ in range(select_steps):
                await self._select(explore=True)

                setups = sampler.sample(group_size)
                awaitables = []
                criticos = []
                for setup in setups:
                    simulator = setup["simulator"]
                    critico = setup["critico"]
                    awaitables.append(
                        simulator.run(
                            num_runs=1,
                            group_id=group_id,
                            agent_version_uuid=agent_version_uuid,
                            environment_uuid=environment_uuid,
                            catch_exception=True,
                        )
                    )
                    criticos.append(critico)

                test_cases_validation, _ = await self._evaluate(
                    awaitables=awaitables, criticos=criticos, verbose=verbose, print_flag="(validation)"
                )

                if pbar is not None:
                    pbar.update(len(test_cases_validation))

                validation_score = 0.0
                for test_case in test_cases_validation:
                    validation_score += test_case["eval_score"]

                    self.versions[self.current_version]["validation_runs"].append(
                        RunSchema(
                            log=test_case["log"],
                            trace_id=test_case["trace_id"],
                            input=test_case["input"],
                            output=test_case["output"],
                            eval_score=test_case["eval_score"],
                            eval_feedback=test_case["eval_feedback"],
                        )
                    )

                    # Kept up to 100 runs per config for browsing
                    self.versions[self.current_version]["validation_runs"] = self.versions[self.current_version][
                        "validation_runs"
                    ][-100:]

                    self.versions[self.current_version]["validation_count"] += 1

                validation_score /= len(test_cases_validation)
                self._update_score(validation_score)

            # Switch to the current version with highest score
            await self._select(explore=False)

            print("\n\n" + "=" * 30 + f" Round {round + 1}/{num_rounds} finishes" + "=" * 30)
            print("Total versions accepted: ", len(self.versions))
            print("Best version index: ", self.current_version)
            print(
                "Score for the best version: %s based on %s rollouts"
                % (
                    self.versions[self.current_version]["average_score"],
                    self.versions[self.current_version]["visits"] * group_size,
                )
            )

            print(
                "All versions: ",
                {
                    i: {
                        "score": self.versions[i]["average_score"],
                        "rollouts evaluated": self.versions[i]["visits"] * group_size,
                    }
                    for i in range(len(self.versions))
                },
            )
            print("--------------------")

            async def sync_to_platform():
                payload = ConfigOptVizSchema(
                    name=self.name,
                    configs=[
                        ConfigSchema(
                            params={
                                key: ParamSchema(**config["params"]._params[key]) for key in config["params"].all()
                            },
                            validation_score=config["average_score"],
                            validation_count=config["validation_count"],
                            validation_runs=config["validation_runs"],
                            utc_timestamp=config["utc_timestamp"],
                        )
                        for config in self.versions
                    ],
                    current_config=self.current_version,
                    validation_score_over_version=[
                        (idx, config["average_score"], config["validation_count"])
                        for idx, config in enumerate(self.versions)
                    ],
                    validation_score_over_time=[
                        (config["utc_timestamp"], config["average_score"], config["validation_count"])
                        for config in self.versions
                    ],
                )

                self.config_opt_viz_id = await self._client.update_config_opt_visual(
                    config_viz=payload, uuid=self.config_opt_viz_id
                )

                return self.config_opt_viz_id

            if self.log_to_platform:
                await sync_to_platform()
                print(
                    (
                        f"Results of round {round + 1}/{num_rounds} uploaded to RELAI platform, "
                        f"visualization id: {self.config_opt_viz_id}\n\n\n"
                    )
                )

        return get_current_params().export()

    async def optimize_structure(
        self,
        total_rollouts: int,
        batch_size: int = 10,
        description: Optional[str] = None,
        code_paths: Optional[list[str]] = None,
        code: Optional[str] = None,
        code_verifier: Optional[Callable[[str], tuple[bool, str]]] = None,
        code_context: Optional[str] = None,
        group_id: str | None = None,
        verbose: bool = True,
        agent_version_uuid: str | None = None,
        environment_uuid: str | None = None,
    ) -> str:
        """
        Propose structural changes (i.e. changes that cannot be achieved by setting parameters alone) to
        improve the agent.

        Args:
            total_rollouts (int): Total number of rollouts to use for optimization.
            batch_size (int): Number of rollouts to use in each batch.
                Generally, a moderate batch_size (e.g. 10-20) is required and recommended.
                For agents with longer execution traces: Try reducing batch_size if an error is raised.
            description (str, optional): Text description of the current structure/workflow/... of the agent.
            code_paths (list[str], optional): A list of paths corresponding to code files containing
                the implementation of the agent.
            code (str, optional): A string representing the code (python, json, etc.) of the agent.
                If both `code_paths` and `code` are provided, `code_paths` will be ignored.
            code_verifier (Callable[[str], tuple[bool, str]], optional):
                A function that verifies the provided code and raises errors if any issues are found.
                It takes a string (code) as input and returns a tuple of a boolean and a string.
                If the boolean is True, the code is verified successfully; otherwise, the string contains
                the issues identified during verification. The verifier will only be applied if `code` or
                `code_paths` is provided, and the optimizer will try to generate code that passes the verification.
            code_context (str, optional): Additional context or information about the code to assist generating
                code that passes the verification. This is effective only when `code_verifier` is provided and is
                especially useful if the code is in json or certain domain-specific languages/formats.
            group_id (str, optional): An optional group ID to associate all runs together. If not provided,
                a new UUID will be generated.
            verbose (bool): If True, additional information will be printed during the optimization.
                Defaults to True.
            agent_version_uuid (str, optional): An optional agent version UUID to associate with the runs.
            environment_uuid (str, optional): An optional environment UUID to associate with the runs.

        Returns:
            str: Suggestion for structural changes to the agent.
        """

        print("\n" + "=" * 30 + "optimize_structure settings" + "=" * 30)
        print("  total_rollouts: ", total_rollouts)
        print("=" * 80 + "\n\n")

        if code is None and code_paths is not None:
            code = extract_code(code_paths=code_paths)

        sampler = ProportionalSampler(
            elements=self.setups,
            weights=[setup["weight"] for setup in self.setups],
        )
        if group_id is None:
            group_id = "Maestro-Struct-" + uuid4().hex

        print("=" * 80)
        print("Running the agent to collect traces...\n\n")

        setups = sampler.sample(total_rollouts)
        awaitables = []
        criticos = []
        for setup in setups:
            simulator = setup["simulator"]
            critico = setup["critico"]
            awaitables.append(
                simulator.run(
                    num_runs=1,
                    group_id=group_id,
                    agent_version_uuid=agent_version_uuid,
                    environment_uuid=environment_uuid,
                    catch_exception=True,
                )
            )
            criticos.append(critico)

        test_cases, _ = await self._evaluate(awaitables=awaitables, criticos=criticos, verbose=verbose)

        if batch_size < len(test_cases):
            print("Selecting and clustering samples for optimization...\n\n")
            active_tags = {}

            for test_case in test_cases:
                _, active_tags = await self._client.process_test_case(
                    {
                        "test_case": test_case,
                        "active_tags": active_tags,
                        "threshold": batch_size * 3,
                        "agent_name": get_full_func_name(self.agent_fn),
                        "agent_code": code,
                        "structure_description": description,
                        "params": get_current_params().export(),
                        "goal": self.goal,
                        "constrained": False,
                    }
                )

            for tag in active_tags:
                active_tags[tag] = 0

            tags_per_test_case = []
            for test_case in test_cases:
                tags, _ = await self._client.process_test_case(
                    {
                        "test_case": test_case,
                        "active_tags": active_tags,
                        "threshold": batch_size * 3,
                        "agent_name": get_full_func_name(self.agent_fn),
                        "agent_code": code,
                        "structure_description": description,
                        "params": get_current_params().export(),
                        "goal": self.goal,
                        "constrained": True,
                    }
                )
                tags_per_test_case.append(tags)
                for tag in tags:
                    if tag in active_tags:
                        active_tags[tag] += 1

            to_sort = []
            for idx in range(len(test_cases)):
                score = 0
                for tag in tags_per_test_case[idx]:
                    score = max(score, active_tags.get(tag, 0))
                to_sort.append((score, idx))
            to_sort.sort(reverse=True)
            selected_test_cases = [test_cases[idx] for score, idx in to_sort[:batch_size]]
        else:
            selected_test_cases = test_cases

        print("=" * 80)
        print("Optimizing structure...\n\n")
        suggestion = await self._client.optimize_structure(
            {
                "agent_name": get_full_func_name(self.agent_fn),
                "agent_code": code,
                "structure_description": description,
                "params": get_current_params().export(),
                "serialized_past_proposals": self._serialize_past_proposals(),
                "test_cases": selected_test_cases,
                "goal": self.goal,
                "max_correction_rounds": 1,
            }
        )

        if (code is not None) and (code_verifier is not None):
            print("Code verifier provided. Generating code...\n\n")
            suggested_code = ""
            verified = False
            issues = None
            correction_round = 0
            max_correction_rounds = 3

            while not verified and correction_round < max_correction_rounds:
                suggested_code = await self._client.apply_structure_to_code(
                    {
                        "agent_name": get_full_func_name(self.agent_fn),
                        "agent_code": code,
                        "suggested_structure": suggestion,
                        "previous_suggested_code": suggested_code,
                        "issues": issues,
                        "code_context": code_context,
                    }
                )
                verified, issues = code_verifier(suggested_code)
                if not verified:
                    print(
                        (
                            f"Code verification failed in round {correction_round + 1}/{max_correction_rounds}\n"
                            f"issues: {issues}\n\n"
                        )
                    )
                correction_round += 1

            if verified:
                print("Code verified successfully!\n\n")
                suggestion_to_upload = (
                    f"[Code verified successfully]\n"
                    f"<proposal>\n{suggestion}\n</proposal>\n\n"
                    f"<suggested_code>\n{suggested_code}\n</suggested_code>\n\n"
                    f"<issues>\n{issues}\n</issues>\n\n"
                )
                suggestion = suggested_code
            else:
                print("Code verification failed.")
                suggestion_to_upload = (
                    f"[Code verification failed]\n"
                    f"<proposal>\n{suggestion}\n</proposal>\n\n"
                    f"<suggested_code>\n{suggested_code}\n</suggested_code>\n\n"
                    f"<issues>\n{issues}\n</issues>\n\n"
                )
                suggestion = suggestion_to_upload
        else:
            suggestion_to_upload = suggestion

        async def sync_to_platform():
            payload = GraphOptVizSchema(
                name=self.name,
                proposal=suggestion_to_upload,
                runs=[
                    RunSchema(
                        log=test_case["log"],
                        trace_id=test_case["trace_id"],
                        input=test_case["input"],
                        output=test_case["output"],
                        eval_score=test_case["eval_score"],
                        eval_feedback=test_case["eval_feedback"],
                    )
                    for test_case in selected_test_cases
                ],
            )

            return await self._client.update_graph_opt_visual(payload)

        print("=" * 40 + "suggestion" + "=" * 40)
        print(suggestion)
        print("=" * 90 + "\n\n")

        if self.log_to_platform:
            uid = await sync_to_platform()
            print(f"Results uploaded to RELAI platform, visualization id: {uid}\n\n\n")

        return suggestion



import json
from collections import defaultdict
from contextlib import contextmanager
from contextvars import ContextVar
from enum import Enum
from typing import Any, Iterator, Mapping, Optional, Union, cast

from openinference.instrumentation import TracerProvider
from openinference.semconv.trace import OpenInferenceSpanKindValues, SpanAttributes, ToolCallAttributes
from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk.trace.export import SimpleSpanProcessor
from opentelemetry.util.types import AttributeValue

from .exporter import InMemorySpanExporter
from .flags import get_current_tracking

exporter = InMemorySpanExporter()
tracer_provider = TracerProvider()
tracer_provider.add_span_processor(SimpleSpanProcessor(exporter))
"""
Tracer provider to be used for openinference instrumentation.
"""


def flatten(mapping: Mapping[str, Any]) -> Iterator[tuple[str, AttributeValue]]:
    for key, value in mapping.items():
        if value is None:
            continue
        if isinstance(value, Mapping):
            for sub_key, sub_value in flatten(value):
                yield f"{key}.{sub_key}", sub_value
        elif isinstance(value, list) and any(isinstance(item, Mapping) for item in value):
            for index, sub_mapping in enumerate(value):
                if isinstance(sub_mapping, Mapping):
                    for sub_key, sub_value in flatten(sub_mapping):
                        yield f"{key}.{index}.{sub_key}", sub_value
                else:
                    yield f"{key}.{index}", sub_mapping
        else:
            if isinstance(value, Enum):
                value = value.value
            yield key, value


def unflatten_attributes(flat: dict[str, Any], prefix: str = "") -> Union[dict[str, Any], list[Any], Any]:
    """
    Convert flattened dot-notation keys into nested dict/list structures.
    Numeric path elements become list indices.
    """
    root: dict[Any, Any] = {}

    for full_key, value in flat.items():
        if prefix:
            if not full_key.startswith(prefix + "."):
                continue
            key = full_key[len(prefix) + 1 :]
        else:
            key = full_key

        parts = key.split(".")
        node = root
        for part in parts[:-1]:
            is_index = part.isdigit()
            idx = int(part) if is_index else part

            if is_index:
                node.setdefault("_list", {})
                node = node["_list"].setdefault(idx, {})
            else:
                node = node.setdefault(part, {})

        leaf = parts[-1]
        if leaf.isdigit():
            idx = int(leaf)
            node.setdefault("_list", {})[idx] = value
        else:
            node[leaf] = value

    def finalize(x: Any) -> Any:
        if isinstance(x, dict):
            if "_list" in x:
                # Turn the stored integer keys into a list
                items = x["_list"]
                return [finalize(items[i]) for i in sorted(items)]
            return {k: finalize(v) for k, v in x.items()}
        return x

    return finalize(root)


def convert_traces(spans: tuple[ReadableSpan, ...], filter_tag: str) -> list[dict]:
    """
    Convert openinference traces to Maestro traces.
    """

    maestro_traces = []

    related_spans = [span for span in spans if filter_tag in span.attributes.get("tag.tags", [])]  # type: ignore

    for span in related_spans:
        attributes = cast(dict[str, Any], span.attributes)
        if attributes.get("openinference.span.kind") == "LLM":
            if attributes.get("relai.kind", "llm") == "llm":
                maestro_traces.append(
                    {
                        "type": "model_calling",
                        "model_name": attributes.get("llm.model_name"),
                        "model_input": attributes.get(
                            "input.value", unflatten_attributes(attributes, "llm.input_messages")
                        ),
                        "model_output": attributes.get(
                            "output.value", unflatten_attributes(attributes, "llm.output_messages")
                        ),
                        "note": attributes.get("relai.note"),
                    }
                )
            elif attributes.get("relai.kind", "llm") == "router":
                maestro_traces.append(
                    {
                        "type": "router",
                        "router_name": attributes.get("llm.model_name"),
                        "router_input": attributes.get(
                            "input.value", unflatten_attributes(attributes, "llm.input_messages")
                        ),
                        "router_output": attributes.get(
                            "output.value", unflatten_attributes(attributes, "llm.output_messages")
                        ),
                        "note": attributes.get("relai.note"),
                    }
                )
        elif attributes.get("openinference.span.kind") == "EMBEDDING":
            "TODO: revisit logging options; further testing required"
        elif attributes.get("openinference.span.kind") == "CHAIN":
            "TODO: revisit logging options; further testing required"
        elif attributes.get("openinference.span.kind") == "RETRIEVER":
            "TODO: revisit logging options; further testing required"
        elif attributes.get("openinference.span.kind") == "RERANKER":
            "TODO: revisit logging options; further testing required"
        elif attributes.get("openinference.span.kind") == "TOOL":
            "TODO: tool_output is not automatically instrumented"
            maestro_traces.append(
                {
                    "type": "tool_calling",
                    "tool_name": attributes.get("tool_call.function.name"),
                    "tool_input": attributes.get("tool_call.function.arguments"),
                    "tool_output": attributes.get("relai.tool_call.output"),
                    "note": attributes.get("relai.note"),
                }
            )
        elif attributes.get("openinference.span.kind") == "AGENT":
            "TODO: revisit logging options; further testing required"

    return maestro_traces


class Logger:
    """
    Logger to gathering execution traces of agents.

    Logs are only collected when `relai_tracking_flag` is enabled (disabled by default and
    can be controlled with `relai.flags.tracking_on` and `relai.flags.tracking_off`).
    """

    def __init__(self, logger_id: str = "default"):
        """
        Initializes the logger with empty logs and tracking structures.
        """
        self._spans = defaultdict(list)
        self._component_log = {}
        self._component_active_params = defaultdict(set)
        self._param_order_store = {}
        self._param_order_counter_store = 0
        self._logger_id = logger_id

    def clear(self):
        """
        Clears all logs and resets tracking state.
        """
        self._spans = defaultdict(list)
        self._component_log = {}
        self._component_active_params = defaultdict(set)
        self._param_order_store = {}
        self._param_order_counter_store = 0
        exporter.clear()

    def log_model(self, name: str, input: Any, output: Any, note: Optional[str] = None):
        """
        Logs a model call event.

        Args:
            name (str): Name of the model.
            input (Any): Input to the model.
            output (Any): Output from the model.
            note (Optional[str]): Optional annotation.
        """
        if get_current_tracking():
            span_id = span_id_var.get()
            self._spans[span_id].append(
                {
                    "type": "model_calling",
                    "model_name": name,
                    "model_input": input,
                    "model_output": output,
                    "note": note,
                }
            )

        tracer = tracer_provider.get_tracer(__name__)
        with tracer.start_as_current_span(name) as span:
            span.set_attribute(SpanAttributes.OPENINFERENCE_SPAN_KIND, OpenInferenceSpanKindValues.LLM.value)
            if isinstance(input, Mapping):
                for attribute, value in flatten(input):
                    span.set_attribute(
                        SpanAttributes.LLM_INPUT_MESSAGES + "." + attribute, value
                    )  # List of messages sent to the LLM in a chat API request, [{"message.role": "user", "message.content": "hello"}]
            else:
                span.set_attribute(SpanAttributes.LLM_INPUT_MESSAGES + ".input", input)
            if isinstance(output, Mapping):
                for attribute, value in flatten(output):
                    span.set_attribute(
                        SpanAttributes.LLM_OUTPUT_MESSAGES + "." + attribute, value
                    )  # Messages received from a chat API, [{"message.role": "user", "message.content": "hello"}]
            else:
                span.set_attribute(SpanAttributes.LLM_OUTPUT_MESSAGES + ".output", output)
            span.set_attribute(SpanAttributes.LLM_MODEL_NAME, name)  # The name of the language model being utilized
            if note is not None:
                span.set_attribute("relai.note", note)  # The note provided by user

    def log_persona(self, name: str, model_name: str, input: Any, output: Any, note: Optional[str] = None):
        """
        Logs a persona activity.

        Args:
            name (str): Name of the persona.
            model_name (str): Name of the model.
            input (Any): Input to the persona.
            output (Any): Output from the persona.
            note (Optional[str]): Optional annotation.
        """
        if get_current_tracking():
            span_id = span_id_var.get()
            self._spans[span_id].append(
                {
                    "type": "persona",
                    "model_name": model_name,
                    "model_input": input,
                    "model_output": output,
                    "note": note,
                }
            )

        tracer = tracer_provider.get_tracer(__name__)
        with tracer.start_as_current_span(name) as span:
            span.set_attribute(SpanAttributes.OPENINFERENCE_SPAN_KIND, OpenInferenceSpanKindValues.LLM.value)
            if isinstance(input, Mapping):
                for attribute, value in flatten(input):
                    span.set_attribute(
                        SpanAttributes.LLM_INPUT_MESSAGES + "." + attribute, value
                    )  # List of messages sent to the LLM in a chat API request, [{"message.role": "user", "message.content": "hello"}]
            else:
                span.set_attribute(SpanAttributes.LLM_INPUT_MESSAGES, input)
            if isinstance(output, Mapping):
                for attribute, value in flatten(output):
                    span.set_attribute(
                        SpanAttributes.LLM_OUTPUT_MESSAGES + "." + attribute, value
                    )  # Messages received from a chat API, [{"message.role": "user", "message.content": "hello"}]
            else:
                span.set_attribute(SpanAttributes.LLM_OUTPUT_MESSAGES, output)
            span.set_attribute(
                SpanAttributes.LLM_MODEL_NAME, model_name
            )  # The name of the language model being utilized
            if note is not None:
                span.set_attribute("relai.note", note)  # The note provided by user

    def log_tool(self, name: str, input: Any, output: Any, note: Optional[str] = None):
        """
        Logs a tool call event.

        Args:
            name (str): Name of the tool.
            input (Any): Input to the tool.
            output (Any): Output from the tool.
            note (Optional[str]): Optional annotation.
        """
        if get_current_tracking():
            span_id = span_id_var.get()
            self._spans[span_id].append(
                {
                    "type": "tool_calling",
                    "tool_name": name,
                    "tool_input": input,
                    "tool_output": output,
                    "note": note,
                }
            )

        tracer = tracer_provider.get_tracer(__name__)
        with tracer.start_as_current_span(name) as span:
            span.set_attribute(SpanAttributes.OPENINFERENCE_SPAN_KIND, OpenInferenceSpanKindValues.TOOL.value)
            span.set_attribute(ToolCallAttributes.TOOL_CALL_FUNCTION_NAME, name)  # The name of the tool being utilized

            if isinstance(input, Mapping):
                for attribute, value in flatten(input):
                    span.set_attribute(
                        ToolCallAttributes.TOOL_CALL_FUNCTION_ARGUMENTS_JSON + "." + attribute, value
                    )  # The arguments for the function being invoked by a tool call
            else:
                span.set_attribute(ToolCallAttributes.TOOL_CALL_FUNCTION_ARGUMENTS_JSON, input)

            if isinstance(output, Mapping):
                for attribute, value in flatten(output):
                    span.set_attribute("relai.tool_call.output" + "." + attribute, value)  # The output of a tool call
            else:
                span.set_attribute("relai.tool_call.output", output)

            if note is not None:
                span.set_attribute("relai.note", note)  # The note provided by user

    def log_router(self, name: str, input: Any, output: Any, note: Optional[str] = None):
        """
        Logs a router event.

        Args:
            name (str): Name of the router.
            input (Any): Input to the router.
            output (Any): Output from the router.
            note (Optional[str]): Optional annotation.
        """
        if get_current_tracking():
            span_id = span_id_var.get()
            self._spans[span_id].append(
                {"type": "router", "router_name": name, "router_input": input, "router_output": output, "note": note}
            )

        # Treating router as a special LLM call in openinference
        tracer = tracer_provider.get_tracer(__name__)
        with tracer.start_as_current_span(name) as span:
            span.set_attribute(SpanAttributes.OPENINFERENCE_SPAN_KIND, OpenInferenceSpanKindValues.LLM.value)

            if isinstance(input, Mapping):
                for attribute, value in flatten(input):
                    span.set_attribute(
                        SpanAttributes.LLM_INPUT_MESSAGES + "." + attribute, value
                    )  # List of messages sent to the LLM in a chat API request, [{"message.role": "user", "message.content": "hello"}]
            else:
                span.set_attribute(SpanAttributes.LLM_INPUT_MESSAGES, input)

            if isinstance(output, Mapping):
                for attribute, value in flatten(output):
                    span.set_attribute(
                        SpanAttributes.LLM_OUTPUT_MESSAGES + "." + attribute, value
                    )  # Messages received from a chat API, [{"message.role": "user", "message.content": "hello"}]
            else:
                span.set_attribute(SpanAttributes.LLM_OUTPUT_MESSAGES, output)

            span.set_attribute(SpanAttributes.LLM_MODEL_NAME, name)  # The name of the language model being utilized
            if note is not None:
                span.set_attribute("relai.note", note)  # The note provided by user
            span.set_attribute("relai.kind", "router")

    def _param_usage(self, param_name: str):
        """
        Records the usage of a parameter during the current span and update
        its ordering index.

        Args:
            param_name (str): Name of the parameter.
        """
        if get_current_tracking():
            span_id = span_id_var.get()
            self._component_active_params[span_id].add(param_name)
            if param_name not in self._param_order_store:
                self._param_order_store[param_name] = self._param_order_counter_store
                self._param_order_counter_store += 1

    def _param_order(self, param_name: str) -> int:
        """
        Returns the ordering index of a parameter.

        Args:
            param_name (str): Parameter name.

        Returns:
            int: Order index of the parameter, or the current counter if not found.
        """
        if param_name not in self._param_order_store:
            return self._param_order_counter_store

        return self._param_order_store[param_name]

    def _component_probe(self, uid: str, input: Any, output: Any):
        """
        Stores input and output logs for a component instance identified by uid.

        Args:
            uid (str): Unique identifier for the component instance.
            input (Any): Input data passed to the component.
            output (Any): Output data returned from the component.
        """
        if get_current_tracking():
            if uid not in self._component_log:
                self._component_log[uid] = []

            self._component_log[uid].append({"input": input, "output": output})

    def component_log(self) -> dict[str, list]:
        """
        Returns all recorded component logs.

        Returns:
            dict[str, list]: Component logs keyed by uid.
        """
        return self._component_log

    def serialize(self):
        """
        Serializes the entire log to a JSON formatted string.

        Returns:
            str: JSON string of all logged events.
        """

        return json.dumps(
            convert_traces(exporter.get_finished_spans(), filter_tag=f"relai.logger.{self._logger_id}"),
            indent=2,
        )

    def to_openinference(self) -> dict[str, list[Any]]:
        """
        Converts the entire log to a list of openinference trace dicts.

        Returns:
            list[dict]: List of all logged events as dicts in openinference format.
        """

        spans = exporter.get_finished_spans()
        return {
            "spans": [
                span.to_json()
                for span in spans
                if f"relai.logger.{self._logger_id}" in span.attributes.get("tag.tags", [])  # type: ignore
            ]
        }


logger = Logger()  # for root context
span_id_var = ContextVar("span_id", default="root")
logger_var: ContextVar[Logger] = ContextVar("logger", default=logger)


@contextmanager
def set_current_logger(logger: Logger):
    """
    A context manager to set the logger for the enclosed context.
    """
    token = logger_var.set(logger)
    try:
        yield
    finally:
        logger_var.reset(token)


def get_current_logger() -> Logger:
    """
    Returns the logger instance for the current execution context.
    """
    return logger_var.get()



from abc import ABC, abstractmethod
from typing import Any


class BaseMocker(ABC):
    def __init__(self) -> None:
        super().__init__()
        self._func_doc: str | None = None
        self._output_type: type | None = None

    @property
    def func_doc(self) -> str | None:
        return self._func_doc

    @func_doc.setter
    def func_doc(self, value: str | None) -> None:
        if self._func_doc is not None:
            raise ValueError("func_doc is already set")
        self._func_doc = value

    @property
    def output_type(self) -> type | None:
        return self._output_type

    @output_type.setter
    def output_type(self, value: type) -> None:
        if self._output_type is not None:
            raise ValueError("output_type is already set")
        self._output_type = value

    @abstractmethod
    def _run(self, *args: Any, **kwargs: Any) -> Any: ...

    async def _arun(self, *args: Any, **kwargs: Any) -> Any:
        return self._run(*args, **kwargs)
